# -*- coding: utf-8 -*-
"""
@date Created on Thur Apr 30 2020
@author martin_g for Eomys
"""

# Standard library imports
import numpy as np


def _main_loudness(spec_third, field_type):
    """Calculate core loudness

    The code is based on BASIC program published in "Program for
    calculating loudness according to DIN 45631 (ISO 532B)", E.Zwicker
    and H.Fastl, J.A.S.J (E) 12, 1 (1991).
    It also corresponds to the following functions of the C program
    published with ISO 532-1:2017:
    - corr_third_octave_intensities
    - f_calc_lcbs
    - f_calc_core_loudness
    - f_corr_loudness

    Parameters
    ----------
    spec_third : numpy.ndarray
        A third octave band spectrum with size (nfreq, nseg) or (nfreq,) [dB ref. 2e-5 Pa]
    field_type : str
        Type of soundfield correspondin to spec_third ("free" or
        "diffuse")

    Outputs
    -------
    nm :  numpy.ndarray
        Core loudness
    """
    # Ref. ISO 532-1:2017 paragraph A.3
    # The table A.3 giving the weights of the 1/3 band levels for
    # center freq. below 300 Hz is only specified for levels up to 120 dB
    # If one of the first 11 bands (from 25 to 250 Hz) exceed 120 dB the
    # Zwicker method cannot be applied.
    if np.max(spec_third[0:11]) > 120.0:
        raise ValueError(
            "1/3 octave band value exceed 120 dB, for which "
            + "the Zwicker method is no longer valid."
        )
    #
    # Date tables definition (variable names and description according to
    # Zwicker:1991)
    # Ranges of 1/3 octave band levels for correction at low frequencies
    # according to equal loudness contours
    rap = np.array([45, 55, 65, 71, 80, 90, 100, 120])
    # Reduction of 1/3 octave band levels at low frequencies according to
    # equal loudness contours within the eight ranges defined by RAP
    dll = np.array(
        [
            (-32, -24, -16, -10, -5, 0, -7, -3, 0, -2, 0),
            (-29, -22, -15, -10, -4, 0, -7, -2, 0, -2, 0),
            (-27, -19, -14, -9, -4, 0, -6, -2, 0, -2, 0),
            (-25, -17, -12, -9, -3, 0, -5, -2, 0, -2, 0),
            (-23, -16, -11, -7, -3, 0, -4, -1, 0, -1, 0),
            (-20, -14, -10, -6, -3, 0, -4, -1, 0, -1, 0),
            (-18, -12, -9, -6, -2, 0, -3, -1, 0, -1, 0),
            (-15, -10, -8, -4, -2, 0, -3, -1, 0, -1, 0),
        ]
    )
    # Critical band level at absolute threshold without taking into
    # account the transmission characteristics of the ear
    ltq = np.array([30, 18, 12, 8, 7, 6, 5, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3])
    # Correction of levels according to the transmission characteristics
    # of the ear
    a0 = np.array(
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, -1.6, -3.2, -5.4, -5.6, -4, -1.5, 2, 5, 12]
    )
    # Level difference between free and diffuse sound fields
    ddf = np.array(
        [
            0,
            0,
            0.5,
            0.9,
            1.2,
            1.6,
            2.3,
            2.8,
            3,
            2,
            0,
            -1.4,
            -2,
            -1.9,
            -1,
            0.5,
            3,
            4,
            4.3,
            4,
        ]
    )
    # Adaptation of 1/3 oct. band levels to the corresponding critical
    # band level
    dcb = np.array(
        [
            -0.25,
            -0.6,
            -0.8,
            -0.8,
            -0.5,
            0,
            0.5,
            1.1,
            1.5,
            1.7,
            1.8,
            1.8,
            1.7,
            1.6,
            1.4,
            1.2,
            0.8,
            0.5,
            0,
            -0.5,
        ]
    )
    #
    # Correction of 1/3 oct. band levels according to equal loudness
    # contours 'xp' and calculation of the intensities for 1/3 oct.
    # bands up to 315 Hz

    # 1-dimensional array to 2-dimensional array with size (nfreq, 1)
    if spec_third.ndim == 1:
        spec_third = spec_third[:, np.newaxis]

    spec_third_aux = spec_third[: dll.shape[1], :]
    # spec_third_aux[:, -1] = 0

    # Convert rap, dll in 3 dimensional array
    # 1. generate the array shape
    base_mat = np.ones((dll.shape[0], dll.shape[1], spec_third_aux.shape[1]))
    # 2. start saving data in rap and DLL array
    rap_mat = np.array(
        [base_mat[:, i, :].T * rap for i in np.arange(dll.shape[1])]
    ).transpose(2, 0, 1)
    dll_mat = np.array(
        [np.multiply(base_mat[:, :, i], dll) for i in np.arange(spec_third.shape[1])]
    ).transpose(1, 2, 0)
    spec_third_aux_mat = np.array(
        [
            np.multiply(base_mat[i, :, :], spec_third_aux)
            for i in np.arange(dll.shape[0])
        ]
    )
    # create the the array rap-dll
    rap_dll_mat = rap_mat - dll_mat

    # This part substitutes the while loop.
    # create the mask to operate
    logic_mat = spec_third_aux_mat > rap_dll_mat
    dll_result = dll_mat[0, :, :]
    dll_result[logic_mat[0, :, :]] = 0
    for i in np.arange(1, dll_mat.shape[0] - 1):
        mask = np.logical_xor(logic_mat[i - 1, :, :], logic_mat[i, :, :])
        dll_result[mask] = dll_mat[i, mask]

    xp = dll_result + spec_third_aux
    ti = np.power(10, (xp / 10))

    # Determination of levels LCB(1), LCB(2) and LCB(3) within the
    # first three critical bands

    gi = np.zeros([3, dll_result.shape[1]])
    gi[0, :] = ti[0:6, :].sum(axis=0)
    gi[1, :] = ti[6:9, :].sum(axis=0)
    gi[2, :] = ti[9:11, :].sum(axis=0)

    logic_gi = gi > 0
    lcb = np.zeros([3, dll_result.shape[1]])
    lcb[logic_gi] = 10 * np.log10(gi[logic_gi])

    # Calculation of main loudness
    s = 0.25
    nm = np.zeros([20, spec_third_aux.shape[1]])
    le = np.copy(spec_third[8:, :])
    # le = le.reshape((20))
    le[0:3, :] = lcb
    a0_mat = np.ones(a0.shape[0] * spec_third.shape[1]).reshape(
        a0.shape[0], spec_third.shape[1]
    )
    a0_mat = (a0_mat.T * a0).T
    le = le - a0_mat

    if field_type == "diffuse":
        ddf_mat = np.ones(ddf.shape[0] * spec_third.shape[1]).reshape(
            ddf.shape[0], spec_third.shape[1]
        )
        ddf_mat = (ddf_mat.T * ddf).T
        le += ddf_mat

    ltq_mat = np.ones(ltq.shape[0] * spec_third.shape[1]).reshape(
        ltq.shape[0], spec_third.shape[1]
    )
    ltq_mat = (ltq_mat.T * ltq).T
    i = le > ltq_mat
    dcb_mat = np.ones(dcb.shape[0] * spec_third.shape[1]).reshape(
        dcb.shape[0], spec_third.shape[1]
    )
    dcb_mat = (dcb_mat.T * dcb).T
    le[i] -= dcb_mat[i]

    mp1 = 0.0635 * np.power(10, 0.025 * ltq_mat)
    mp1[i == False] = 0
    mp2 = np.power(1 - s + s * np.power(10, 0.1 * (le - ltq_mat)), 0.25) - 1
    mp2[i == False] = 0

    nm = np.multiply(mp1, mp2)

    nm[i == False] = 0
    nm[nm < 0] = 0
    current_shape = nm.shape
    nm = np.append(nm, np.zeros(current_shape[1])).reshape(
        current_shape[0] + 1, current_shape[1]
    )
    #
    # Correction of specific loudness in the lowest critical band
    # taking into account the dependance of absolute threshold
    # within this critical band
    korry = 0.4 + 0.32 * nm[0] ** 0.2
    nm[0, korry <= 1] *= korry[korry <= 1]
    # nm[:, -1] = 0
    # if spec_third.ndim == 1 or spec_third.shape[1] == 1:
    #     # This is only for test_loudness_zwicker_3oct because only one array of one col is given and this routine needs 2 or more
    #     # This line is only also for testing test_loudness_zwicker_wav(), only in case one col in spec third is given.
    #     nm = nm[:, 1]

    return np.squeeze(nm)



# -*- coding: utf-8 -*-

# Standard imports
import numpy as np

# Local imports
from mosqito.utils.time_segmentation import time_segmentation
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sq_metrics.roughness.roughness_dw._roughness_dw_main_calc import (
    _roughness_dw_main_calc,
)
from mosqito.sq_metrics.roughness.roughness_dw._gzi_weighting import _gzi_weighting
from mosqito.sq_metrics.roughness.roughness_dw._H_weighting import _H_weighting

# Optional package import
try:
    from SciDataTool import DataTime, DataLinspace, DataFreq, Norm_func
except ImportError:
    DataTime = None
    DataLinspace = None
    DataFreq = None


def roughness_dw(signal, fs=None, overlap=0.5, is_sdt_output=False):
    """Roughness calculation of a signal sampled at 48kHz.

    The code is based on the algorithm described in "Psychoacoustical roughness:
    implementation of an optimized model" by Daniel and Weber in 1997.
    The roughness model consists of a parallel processing structure that is made up
    of successive stages and calculates intermediate specific roughnesses R_spec,
    which are summed up to determine the total roughness R.

    Parameters
    ----------
    signal :numpy.array  or DataTime object
        A time signal in Pa
    fs : float, optional
        Sampling frequency, can be omitted if the input is a DataTime
        object. Default to None
    overlap : float
        Overlapping coefficient for the time windows of 200ms

    Outputs
    -------
    R : numpy.array
        Roughness in [asper].
    R_spec : numpy.array
        Specific roughness over bark axis.
    bark_axis : numpy.array
        Frequency axis in [bark].
    time : numpy.array
        Time axis in [s].

    """

    # Manage input type
    if DataTime is not None and isinstance(signal, DataTime):
        time = signal.get_along("time")["time"]
        fs = 1 / (time[1] - time[0])
        signal = signal.get_along("time")[signal.symbol]

    # Number of points within each frame according to the time resolution of 200ms
    nperseg = int(0.2 * fs)
    # Overlapping segment length
    noverlap = int(overlap * nperseg)
    # reshaping of the signal according to the overlap and time proportions
    sig, time = time_segmentation(
        signal, fs, nperseg=nperseg, noverlap=noverlap, is_ecma=False
    )
    if len(sig.shape) == 1:
        nseg = 1
    else:
        nseg = sig.shape[1]

    spec, _ = spectrum(sig, fs, nfft="default", window="blackman", db=False)

    # Frequency axis in Hertz
    freq_axis = np.arange(1, nperseg // 2 + 1, 1) * (fs / nperseg)

    # Initialization of the weighting functions H and g
    hWeight = _H_weighting(nperseg, fs)
    # Aures modulation depth weighting function
    gzi = _gzi_weighting(np.arange(1, 48, 1) / 2)

    R = np.zeros((nseg))
    R_spec = np.zeros((47, nseg))
    if len(spec.shape) > 1:
        for i in range(nseg):
            R[i], R_spec[:, i], bark_axis = _roughness_dw_main_calc(
                spec[:, i], freq_axis, fs, gzi, hWeight
            )
    else:
        R, R_spec, bark_axis = _roughness_dw_main_calc(
            spec, freq_axis, fs, gzi, hWeight
        )

    # print(np.mean(R,axis=0))

    # Manage SciDataTool output type
    if is_sdt_output:
        if DataLinspace is None:
            raise RuntimeError(
                "In order to handle Data objects you need the 'SciDataTool' package."
            )
        else:
            bark_data = DataLinspace(
                name="Critical band rate",
                unit="Bark",
                initial=bark_axis[0],
                final=bark_axis[-1],
                number=len(bark_axis),
                include_endpoint=True,
                normalizations={
                    "Hz": Norm_func(function=lambda x: 1960 * (x + 0.53) / (26.28 - x))
                },
            )
            time = DataLinspace(
                name="time",
                unit="s",
                initial=time[0],
                final=time[-1],
                number=len(time),
                include_endpoint=True,
            )
            R_spec = DataFreq(
                name="Specific roughness (Daniel & Weber method)",
                symbol="R'_{dw}",
                axes=[bark_data, time],
                values=R_spec,
                unit="asper/Bark",
            )
            R = DataTime(
                name="Roughness (Daniel & Weber method)",
                symbol="R_{dw}",
                axes=[time],
                values=R,
                unit="asper",
            )

    return R, R_spec, bark_axis, time



# -*- coding: utf-8 -*-

# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError(
        "In order to perform the tests you need the 'pytest' package."
    )

import numpy as np
# Local application imports
from mosqito.utils import load
from mosqito.sq_metrics import pr_ecma_tv


@pytest.mark.pr_tv  # to skip or run PR test
def test_pr_ecma_tv():
    """Test function for the prominence ratio calculation of an audio signal

    Validation function for the Audio_signal class "tone_to_noise_ecma" method with signal array
    as input. The input signals are generated using audacity.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    # Test signal as input for prominence ratio calculation
    # signals generated using audacity : white noise + tones at 442 and 1768 Hz

    signal = {
            "data_file": "tests/input/white_noise_442_1768_Hz_varying.wav"
        }

    # Load signal
    audio, fs = load(signal["data_file"], wav_calib=0.01)

    # Compute tone-to-noise ratio
    t_pr, pr, prom, freq, time = pr_ecma_tv(audio, fs, prominence=True)
    np.testing.assert_almost_equal(max(t_pr), 34.02082433185258)
    assert pr[np.argmin(np.abs(freq-442)),:].all() != np.nan
    assert pr[np.argmin(np.abs(freq-1768)),2:3].all() != np.nan
    assert np.count_nonzero(prom == True) == 8



if __name__ == "__main__":
    test_pr_ecma_tv()





# -*- coding: utf-8 -*-

import numpy as np

# from scipy.signal import welch, periodogram

# Local functions imports
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._critical_band import (
    _critical_band,
    _lower_critical_band,
    _upper_critical_band,
)
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._screening_for_tones import _screening_for_tones
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._find_highest_tone import _find_highest_tone


def _pr_main_calc(spectrum_db, freq_axis):
    """
        Calculation of the tone-to noise ratio according to the method described
        in ECMA 74, annex D.
        The method to find the tonal candidates is the one described by Wade Bray
        in 'Methods for automating prominent tone evaluation and for considerin
        variations with time or other reference quantities'
        The total value calculated, T-PR, is calculated according to ECMA TR/108

    Parameters
    ----------
    spectrum_db : array
        Spectrum values in dB [nperseg x nseg]
    freq_axis : array
        frequency axis corresponding to the spectrum [nperseg x nseg]

    Output
    ------
    tones_freqs : array of float
        Frequency list of the tones [number of tones x nseg]
    pr : array of float
        TNR value calculated for each tone [number of tones x nseg]
    prominence : array of boolean
        Prominence criteria as described in ECMA 74 [number of tones x nseg]
    t_pr : array of float
        Sum of the specific TNR
    """

    #### Spectrum creation #######################################################

    if len(spectrum_db.shape) == 1:
        nseg = 1
        # Frequency axis of interest
        freq_index = np.where((freq_axis > 89.1) & (freq_axis < 11200))[0]
        freqs = freq_axis[freq_index]
        spec_db = spectrum_db[freq_index]

    elif (len(spectrum_db.shape) > 1) & (len(freq_axis.shape) > 1):
        nseg = spectrum_db.shape[1]
        freqs = [[]for i in range(nseg)]
        spec_db = [[]for i in range(nseg)]
        for i in range(nseg):
            # Frequency axis of interest
            freq_index_rows = np.where(
                (freq_axis[:, i] > 89.1) & (freq_axis[:, i] < 11200))[0]
            freqs[i] = np.append(freqs[i], freq_axis[freq_index_rows, i])
            spec_db[i] = np.append(spec_db[i], spectrum_db[freq_index_rows, i])
        freqs = np.asarray(freqs)
        spec_db = np.asarray(spec_db)

    elif (len(spectrum_db.shape) > 1) & (len(freq_axis.shape) == 1):
        # Frequency axis of interest
        freq_index = np.where((freq_axis > 89.1) & (freq_axis < 11200))[0]
        # Initialization
        nfreqs = len(freq_index)
        nseg = spectrum_db.shape[1]
        freqs = np.zeros((nseg, nfreqs))
        spec_db = np.zeros((nseg, nfreqs))
        for i in range(nseg):
            freqs[i, :] = freq_axis[freq_index]
            spec_db[i, :] = spectrum_db[freq_index, i]

    #### Screening to find the potential tonal components ########################

    peak_index = _screening_for_tones(freqs, spec_db, "smoothed", 90, 11200)

    #### Evaluation of each candidate ############################################

    # Initialization of the results lists
    if nseg == 1:
        PR = []
        t_pr = []
        tones_freqs = []
        prominence = []
    else:
        PR = [[]for i in range(nseg)]
        t_pr = [[]for i in range(nseg)]
        tones_freqs = [[]for i in range(nseg)]
        prominence = [[]for i in range(nseg)]

    for i in range(nseg):

        pr = np.array([], dtype=object)

        if nseg == 1:
            peaks = peak_index
            spec = spec_db
            fr = freqs

        elif nseg > 1:
            peaks = peak_index[i]
            spec = spec_db[i, :]
            fr = freqs[i, :]

        nb_tones = len(peaks)

        # Each candidate is studied and then deleted from the list until all have been treated
        while nb_tones > 0:
            ind = int(peaks[0])

            # Find the highest tone in the critical band
            if len(peaks) > 1:
                ind, _, peaks, nb_tones = _find_highest_tone(
                    fr, spec, peaks.astype(int), nb_tones, ind
                )

            ft = fr[ind]

            # Level of the middle critical band
            f1, f2 = _critical_band(ft)
            low_limit_idx = np.argmin(np.abs(fr - f1))
            high_limit_idx = np.argmin(np.abs(fr - f2))

            spec_sum = sum(10 ** (spec[low_limit_idx:high_limit_idx] / 10))
            if spec_sum != 0:
                Lm = 10 * np.log10(spec_sum)
            else:
                Lm = 0

            # Level of the lower critical band
            f1, f2 = _lower_critical_band(ft)
            low_limit = np.argmin(np.abs(fr - f1))
            high_limit = np.argmin(np.abs(fr - f2))

            spec_sum = sum(10 ** (spec[low_limit:high_limit] / 10))
            if spec_sum != 0:
                Ll = 10 * np.log10(spec_sum)
            else:
                Ll = 0

            delta_f = f2 - f1

            # Level of the upper critical band
            f1, f2 = _upper_critical_band(ft)
            low_limit = np.argmin(np.abs(fr - f1))
            high_limit = np.argmin(np.abs(fr - f2))

            spec_sum = sum(10 ** (spec[low_limit:high_limit] / 10))
            if spec_sum != 0:
                Lu = 10 * np.log10(spec_sum)
            else:
                Lu = 0

            if ft <= 171.4:
                delta = 10 * np.log10(10 ** (0.1 * Lm)) - 10 * np.log10(
                    ((100 / delta_f) * 10 ** (0.1 * Ll) + 10 ** (0.1 * Lu)) * 0.5
                )

            elif ft > 171.4:
                delta = 10 * np.log10(10 ** (0.1 * Lm)) - 10 * np.log10(
                    (10 ** (0.1 * Ll) + 10 ** (0.1 * Lu)) * 0.5
                )

            if delta > 0:
                if nseg > 1:
                    tones_freqs[i] = np.append(tones_freqs[i], ft)
                elif nseg == 1:
                    tones_freqs = np.append(tones_freqs, ft)
                pr = np.append(pr, delta)

                # Prominent discrete tone criteria
                if ft >= 89.1 and ft <= 1000:
                    if delta >= 9 + 10 * np.log10(1000 / ft):
                        if nseg > 1:
                            prominence[i].append(True)

                        elif nseg == 1:
                            prominence.append(True)
                    else:
                        if nseg > 1:
                            prominence[i].append(False)

                        elif nseg == 1:
                            prominence.append(False)

                elif ft > 1000:
                    if delta >= 9:
                        if nseg > 1:
                            prominence[i].append(True)

                        elif nseg == 1:
                            prominence.append(True)
                    else:
                        if nseg > 1:
                            prominence[i].append(False)

                        elif nseg == 1:
                            prominence.append(False)

            # suppression from the list of tones of all the candidates belonging to the
            # same critical band
            sup = np.where((peaks >= low_limit_idx) & (peaks <= high_limit_idx))[0]
            peaks = np.delete(peaks, sup)
            nb_tones -= len(sup)

        if nseg > 1:
            if sum(np.power(10, (pr[prominence[i]] / 10))) != 0:
                t_pr[i] = 10 * \
                    np.log10(sum(np.power(10, (pr[prominence[i]] / 10))))
            else:
                t_pr[i] = 0
            PR[i] = np.append(PR[i], pr)

        elif nseg == 1:
            if sum(np.power(10, (pr[prominence] / 10))) != 0:
                t_pr = np.append(
                    t_pr, 10 * np.log10(sum(np.power(10, (pr[prominence] / 10)))))
            else:
                t_pr = 0
            PR = np.append(PR, pr)

    tones_freqs = np.asarray(tones_freqs, dtype=object)
    prominence = np.asarray(prominence, dtype=object)

    return tones_freqs, PR, prominence, t_pr



# -*- coding: utf-8 -*-

import numpy as np
import math


def sine_wave_generator(fs, t, spl_value, freq):
    """It creates a sine wave signal given some input parameters like frequency, duration, sampling rate or sound
    pressure level.

    Parameters
    ----------
    fs: int
        'Hz', sampling frequency.
    t: int
        's', signal duration.
    spl_value: int
        'dB SPL', signal sound pressure level.
    freq: int
        'Hz', sine wave frequency.

    Returns
    -------
    signal: numpy.array
        'Pa', time signal values. For ECMA-418-2 the sampling frequency of the signal must be 48000 Hz.
    time: numpy.array
        Time scale arranged in a numpy array.
    """

    # "Peak" value in Pascals (amplitude)
    p_ref = 2e-5
    pressure_rms = p_ref * (10.00 ** (spl_value / 20.00))

    # Sample range
    # samples = np.linspace(0, t, int(fs * t), endpoint=False)
    time = np.arange(0, t, 1 / fs)

    # Theta lets you specify the sine wave value at time 0
    theta = 0

    # Amplitude of the signal
    amplitude = np.sqrt(2) * pressure_rms

    # Signal calculation
    signal = np.array(amplitude * np.sin((2.00 * math.pi * freq * time) + theta))

    return signal, time



# -*- coding: utf-8 -*-

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError(
        "In order to perform this validation you need the 'matplotlib' package."
    )

import numpy as np


def isoclose(actual, desired, rtol=1e-7, atol=0, is_plot=False, tol_label=None, xaxis=None):
    """
    Check if two arrays are equal up to desired tolerance.

    The test is inspired from section 5.1 of ISO 532-1. It compares 
    ``actual`` to ``desired +/- min(atol, rtol * abs(desired))``.

    Parameters
    ----------
    actual : array_like
        Array obtained.
    desired : array_like
        Array desired.
    rtol : float, optional
        Relative tolerance.
    atol : float, optional
        Absolute tolerance.
    is_plot : bool, optional
        To generate a "compliance" plot
    tol_label: str
        Label for the tolerance curves
    xaxis : array_like, optional
        x axis for the "compliance" plot

    Output
    ------
    is_isoclose: bool
        False if actual and desired are not equal up to specified tolerance.

    """

    # Tolerances
    range_pos = np.amin(
        [desired * (1 - abs(rtol)), desired - abs(atol)], axis=0)
    range_neg = np.amax(
        [desired * (1 + abs(rtol)), desired + abs(atol)], axis=0)

    # Test for ISO 532-1 comformance (section 5.1)
    is_isoclose = (actual >= range_pos).all() and (actual <= range_neg).all()

    if is_plot:
        # Define xaxis
        if xaxis is None:
            xaxis = np.arange(actual.shape[0])

        # Plot desired range
        plt.plot(
            xaxis,
            range_neg,
            color="tab:red",
            linestyle="solid",
            label=tol_label,
            linewidth=1,
        )
        plt.plot(
            xaxis,
            range_pos,
            color="tab:red",
            linestyle="solid",
            # label="",
            linewidth=1,
        )

        # Plot actual value
        plt.plot(xaxis, actual)
        plt.legend()

    return is_isoclose



# -*- coding: utf-8 -*-

# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError(
        "In order to perform the tests you need the 'pytest' package."
    )
try:
    from SciDataTool import DataLinspace, DataTime
except ImportError:
    raise RuntimeError(
        "In order to handle Data objects you need the 'SciDataTool' package."
    )

import numpy as np
from scipy.fft import fft

# Local application imports
from mosqito.sq_metrics import roughness_dw, roughness_dw_freq
from tests.sq_metrics.roughness.signals_test_generation import signal_test
from mosqito.sound_level_meter.spectrum import spectrum

@pytest.mark.roughness_dw  # to skip or run only Daniel and Weber roughness tests
def test_roughness_dw():
    """Test function for the roughness calculation of a audio signal

    Test function for the script "comp_roughness" method with signal array
    as input. The input signals are chosen according to the article "Psychoacoustical
    roughness: implementation of an optimized model" by Daniel and Weber in 1997.
    The figure 3 is used to compare amplitude-modulated signals created according to
    their carrier frequency and modulation frequency to the article results.
    One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """

    # Stimulus generation
    stimulus, _ = signal_test(
        fc=1000, fmod=70, mdepth=1, fs=44100, d=0.2, dB=60)

    # Roughness calculation
    roughness, time, _, _ = roughness_dw(stimulus, fs=44100, overlap=0)
    R = {
        "name": "Roughness",
        "values": roughness,
        "time": time,
    }

    # Check compliance
    tst = check_compliance(R)

    assert tst


@pytest.mark.roughness_dw  # to skip or run only Daniel and Weber roughness tests
def test_roughness_dw_sdt():
    """Test function for the roughness calculation of a audio signal

    Test function for the script "comp_roughness" method with signal array
    as input. The input signals are chosen according to the article "Psychoacoustical
    roughness: implementation of an optimized model" by Daniel and Weber in 1997.
    The figure 3 is used to compare amplitude-modulated signals created according to
    their carrier frequency and modulation frequency to the article results.
    One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """

    # Stimulus generation
    stimulus, time = signal_test(
        fc=1000, fmod=70, mdepth=1, fs=44100, d=0.2, dB=60)
    time = DataLinspace(
        name="time",
        unit="s",
        initial=0,
        final=(len(time) - 1) / 44100,
        number=len(time),
        include_endpoint=True,
    )
    stimulus = DataTime(
        name="Test signal",
        symbol="p",
        unit="Pa",
        axes=[time],
        values=stimulus,
    )
    # Roughness calculation
    roughness, time, _, _ = roughness_dw(
        stimulus, fs=44100, overlap=0, is_sdt_output=True)
    roughness = roughness.get_along('time')[roughness.symbol]

    R = {
        "name": "Roughness",
        "values": roughness,
        "time": time,
    }

    # Check compliance
    tst = check_compliance(R)

    assert tst


@pytest.mark.roughness_dw  # to skip or run only Daniel and Weber roughness tests
def test_roughness_dw_freq():
    """Test function for the roughness calculation of a audio signal

    Test function for the script "comp_roughness" method with spectrum array
    as input. The input signals are chosen according to the article "Psychoacoustical
    roughness: implementation of an optimized model" by Daniel and Weber in 1997.
    The figure 3 is used to compare amplitude-modulated signals created according to
    their carrier frequency and modulation frequency to the article results.
    One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    fs = 44100
    # Stimulus generation
    stimulus, _ = signal_test(
        fc=1000, fmod=70, mdepth=1, fs=fs, d=0.2, dB=60)

    # conversion into frequency domain
    n = len(stimulus)
    spec, freqs = spectrum(stimulus, fs, nfft="default", window="blackman", db=False)

    # Roughness calculation
    roughness, _, _ = roughness_dw_freq(spec, freqs)
    R = {
        "name": "Roughness",
        "values": roughness,
    }
    print(R)

    # Check compliance
    tst = check_compliance(R)

    assert tst


def check_compliance(R):
    """Check the compliance of roughness calc. to Daniel and Weber article
    "Psychoacoustical roughness: implementation of an optimized model", 1997.

    Check the compliance of the input data R to figure 3 of the article
    using the reference data described in the dictionary article_ref.

    Parameter
    ---------
    R: numpy.array
        Calculated roughnesses [asper]

    Output
    ------
    tst : bool
        Compliance to the reference data
    """

    # Reference value of 1 asper given by Zwicker and Fastl
    ref = 1

    # Test for comformance (17% tolerance)
    tst = (R["values"] >= ref * 0.83).all() and (R["values"] <= ref * 1.17).all()

    return tst


# test de la fonction
if __name__ == "__main__":
    test_roughness_dw()
    test_roughness_dw_sdt()
    test_roughness_dw_freq()



# -*- coding: utf-8 -*-

# Standard library import
import numpy as np

# Local functions imports
from mosqito.utils.time_segmentation import time_segmentation
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._tnr_main_calc import _tnr_main_calc


def tnr_ecma_tv(signal, fs, prominence=False, overlap=0):
    """Computation of tone-to-noise ration according to ECMA-74, annex D.9
    for a time varying signal.
        The T-TNR value is calculated according to ECMA-TR/108

    Parameters
    ----------
    signal :numpy.array
        A time varying signal in [Pa].
    fs : integer
        Sampling frequency.
    prominence : boolean
        If True, the algorithm only returns the prominent tones, if False it returns all tones detected.
        Default is True.
    overlap : float
        Overlapping parameter for the time frames of 500ms. Default is 0.5.

    Output
    ------
    t_tnr : array of float
        global TNR value, along time if is_stationary = False
    tnr : array of float
        TNR values for each detected tone
    promi : array of bool
        prominence criterion for each detected tone
    tones_freqs : array of float
        frequency of the detected tones
    time  : array of float, only if is_stationary = False
        time axis
    """
    
    if len(signal.shape) == 1:
      
        # Number of points within each frame according to the time resolution of 500ms
        nperseg = int(0.5 * fs)
        # Overlappinf segment length
        noverlap = int(overlap * nperseg)               
        # Time segmentation of the signal
        sig, time = time_segmentation(signal, fs, nperseg=nperseg, noverlap=noverlap, is_ecma=False)
        # Number of segments
        nseg = sig.shape[1] 
        # Spectrum computation
        spectrum_db, freq_axis = spectrum(sig, fs, db=True)
      
    else:
        nseg = signal.shape[1]
        time = np.linspace(0, signal.shape[0]/fs, num=nseg)
        
        # Compute spectrum
        spectrum_db, freq_axis = spectrum(sig, fs, db=True)
            
            
    # compute tnr values
    tones_freqs, tnr_, prom, t_tnr = _tnr_main_calc(spectrum_db, freq_axis)
 
            
    # Retore the results in a time vs frequency array
    freqs = np.logspace(np.log10(90), np.log10(11200), num=1000)
    tnr = np.empty((len(freqs), nseg))
    tnr.fill(np.nan)
    promi = np.empty((len(freqs), nseg), dtype=bool)
    promi.fill(False)
    
    for t in range(nseg):
        for f in range(len(tones_freqs[t])):
            ind = np.argmin(np.abs(freqs - tones_freqs[t][f]))
            if prominence == False:
                tnr[ind, t] = tnr_[t][f]
                promi[ind, t] = prom[t][f]
            if prominence == True:
                if prom[t][f] == True:
                    tnr[ind, t] = tnr_[t][f]
                    promi[ind, t] = prom[t][f]

    t_tnr = np.ravel(t_tnr)

    return t_tnr, tnr, promi, freqs, time     



# -*- coding: utf-8 -*-


# Standard library imports
import numpy as np
from scipy.signal import butter, sosfreqz

def _n_oct_freq_filter(spectrum, fs, fc, alpha, n=3):  
    """ n-th octave filtering in frequency domain

    Designs a digital 1/3-octave filter with center frequency fc for
    sampling frequency fs. 

    Parameters
    ----------
    spec : numpy.ndarray
        Frequency spectrum [complex]
    fs : float
        Sampling frequency [Hz]
    fc : float
        Filter exact center frequency [Hz]
    alpha : float
        Ratio of the upper and lower band-edge frequencies to the mid-band
        frequency
    n : int, optional
        Filter order. Default to 3

    Outputs
    -------
    level : float
        Rms level of sig in the third octave band centered on fc
    """

    # Normalized cutoff frequencies
    w1 = fc / (fs  / 2) / alpha
    w2 = fc / (fs  / 2) * alpha

    # Define filter coefficient
    sos = butter(n, [w1, w2], "bandpass", analog=False, output ='sos')  
    # Get FRF and apply it
    w, h = sosfreqz(sos, worN=len(spectrum))
    spec_filt = np.multiply(h, spectrum.T).T
    
    # Compute overall rms level
    level = np.sqrt(np.sum(np.abs(spec_filt) ** 2, axis=0)) 
        
    return level



# -*- coding: utf-8 -*-

# Local imports
from mosqito.sq_metrics import loudness_zwst
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_from_loudness import sharpness_din_from_loudness


def sharpness_din_st(signal, fs, weighting="din", field_type="free"):
    """Acoustic sharpness calculation according to different methods:
        Aures, Von Bismarck, DIN 45692, Fastl

    Parameters:
    ----------
    signal: numpy.array
        A time signal in [Pa], dim (nperseg, nseg)
    fs: integer
        Sampling frequency
    weighting : string
        To specify the weighting function used for the
        sharpness computation.'din' by default,'aures', 'bismarck','fastl'
    field_type : str
        Type of soundfield corresponding to spec_third ("free" by
        default or "diffuse").

    Outputs
    ------
    S : float
        Sharpness value, dim (nseg)

    """

    # Compute loudness
    N, N_specific, _ = loudness_zwst(signal, fs, field_type=field_type)

    # Compute sharpness from loudness
    S = sharpness_din_from_loudness(N, N_specific, weighting=weighting)

    return S



# -*- coding: utf-8 -*-

# External import
import numpy as np
# Local functions imports
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._tnr_main_calc import _tnr_main_calc
from mosqito.utils.conversion import amp2db

def tnr_ecma_freq(spectrum, freqs,  prominence=True):
    """Computation of tone-to-noise ration according to ECMA-74, annex D.9
    for an amplitude or complex spectrum.
        The T-TNR value is calculated according to ECMA-TR/108

    Parameters
    ----------
    spectrum :numpy.array
        Amplitude or complex frequency spectrum [nperseg x nseg].
    freqs : np.array
        Frequency axis [nperseg x nseg] or [nperseg]. 
    prominence : boolean
        If True, the algorithm only returns the prominent tones, if False it returns all tones detected.
        Default is True.

    Output
    ------
    t_tnr : array of float
        Global TNR value.
    tnr : array of float
        TNR values for each detected tone.
    promi : array of bool
        Prominence criterion for each detected tone.
    tones_freqs : array of float
        Frequency list of the detected tones.
    """
             
    if len(spectrum) != len(freqs) :
        raise ValueError('Input spectrum and frequency axis must have the same size')
    
    # Compute spectrum dB values
    spectrum_db = amp2db(np.abs(spectrum), ref=2e-5)
            
    # compute TNR values
    tones_freqs, tnr, prom, t_tnr = _tnr_main_calc(spectrum_db, freqs)
  
    if prominence == False:
        return t_tnr, tnr, prom, tones_freqs
    else:
        return t_tnr, tnr[prom], prom[prom], tones_freqs[prom]



# -*- coding: utf-8 -*-

import numpy as np

# Local functions imports
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._critical_band import _critical_band
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._screening_for_tones import _screening_for_tones
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._find_highest_tone import _find_highest_tone
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._peak_level import _peak_level


def _tnr_main_calc(spectrum_db, freq_axis):
    """
        Calculation of the tone-to noise ratio according to the method described
        in ECMA 74, annex D.
        The method to find the tonal candidates is the one described by Wade Bray
        in 'Methods for automating prominent tone evaluation and for considerin
        variations with time or other reference quantities'
        The total value calculated, T-TNR, is calculated according to ECMA TR/108

    Parameters
    ----------
    spectrum_db : array
        spectrum values in dB [nperseg x nseg].
    freq_axis : array
        frequency axis corresponding to the spectrum [nperseg x nseg].

    Output
    ------
    tones_freqs : array of float
        frequency of the tones (n blocks x number of tones)
    tnr : array of float
        TNR value calculated for each tone (n blocks x number of tones)
    prominence : array of boolean
        prominence criteria as described in ECMA 74 (n blocks x number of tones)
    t_tnr : array of float
        sum of the specific TNR (n blocks x global value)
    """

    #### Spectrum creation #######################################################


    if len(spectrum_db.shape) == 1:
        nseg = 1
        # Frequency axis of interest
        freq_index = np.where((freq_axis > 89.1) & (freq_axis < 11200))[0]
        freqs = freq_axis[freq_index]
        spec_db = spectrum_db[freq_index]
        
    elif (len(spectrum_db.shape) > 1) & (len(freq_axis.shape) > 1):
        nseg = spectrum_db.shape[1]
        freqs = [[]for i in range(nseg)]
        spec_db = [[]for i in range(nseg)]
        for i in range(nseg):
            # Frequency axis of interest
            freq_index_rows = np.where((freq_axis[:,i] > 89.1) & (freq_axis[:,i] < 11200))[0]
            freqs[i] = np.append(freqs[i],freq_axis[freq_index_rows,i])
            spec_db[i] = np.append(spec_db[i],spectrum_db[freq_index_rows,i])
        freqs = np.asarray(freqs)
        spec_db = np.asarray(spec_db)
    
    elif (len(spectrum_db.shape) > 1) & (len(freq_axis.shape) == 1):
        # Frequency axis of interest
        freq_index = np.where((freq_axis > 89.1) & (freq_axis < 11200))[0]
        # Initialization
        nfreqs = len(freq_index)    
        nseg = spectrum_db.shape[1]
        freqs = np.zeros((nseg,nfreqs))
        spec_db = np.zeros((nseg,nfreqs))
        for i in range(nseg):
            freqs[i,:] = freq_axis[freq_index]
            spec_db[i,:] = spectrum_db[freq_index,i]


    #### Screening to find the potential tonal components ########################

    peak_index = _screening_for_tones(freqs, spec_db, "smoothed", 90, 11200)

    # Initialization of the results lists
    if nseg == 1:
        TNR = []
        t_tnr = []
        tones_freqs = []
        prominence = []
    else:   
        TNR = [[]for i in range(nseg)]
        t_tnr = [[]for i in range(nseg)]
        tones_freqs = [[]for i in range(nseg)]
        prominence = [[]for i in range(nseg)]
    

    #### Evaluation of each candidate ############################################

    for i in range(nseg):
        
        tnr = np.array([], dtype=object)
        
        if nseg == 1:
            peaks = peak_index
            spec = spec_db
            fr = freqs
            frs = freq_axis
        elif nseg > 1:
            peaks = peak_index[i]
            spec = spec_db[i,:]
            fr = freqs[i,:]
            if len(freq_axis.shape)>1:
                frs = freq_axis[:,i]
            else:
                frs = freq_axis
        
        nb_tones = len(peaks)

        # Each candidate is studied and then deleted from the list until all have been treated
        while nb_tones > 0:
            ind = int(peaks[0])
            if len(peaks) > 1:
                ind_p, ind_s, peaks, nb_tones = _find_highest_tone(
                    fr, spec, peaks.astype(int), nb_tones, ind
                )
            else:
                ind_p = ind
                ind_s = None
    
            # multiple tones in a critical band
            if ind_s != None:
                fp = fr[ind_p]
                fs = fr[ind_s]
    
                # proximity criterion
                delta_f = 21 * 10 ** ((1.2 * (np.abs(np.log10(fp / 212))) ** 1.8))
                if np.abs(fs - fp) < delta_f:
    
                    # tone SPL
                    Lp = _peak_level(fr, spec, ind_p)
                    Ls = _peak_level(fr, spec, ind_s)
    
                    Lt = 10 * np.log10(((10 ** (Lp / 10) + 10 ** (Ls / 10))))
    
                    # total SPL in the critical band
                    f1, f2 = _critical_band(fp)
                    low_limit_idx = np.argmin(np.abs(fr - f1))
                    high_limit_idx = np.argmin(np.abs(fr - f2))
    
                    spec_sum = sum(10 ** (spec[low_limit_idx:high_limit_idx] / 10))
                    Ltot = 10 * np.log10(spec_sum)
    
                    # suppression of the second highest tone from the list of tones
                    sup = np.where(peaks == ind_s)[0]
                    peaks = np.delete(peaks, sup)
                    nb_tones -= 1
    
                    delta_ft = 2 * (frs[1] - frs[0])
    
                else:
                    # the two highest tones are not close enough to be considered as one
                    # tone SPL
                    Lt = spec[ind_p]
    
                    # total SPL in the critical band
                    f1, f2 = _critical_band(fr[ind_p])
                    low_limit_idx = np.argmin(np.abs(fr - f1))
                    high_limit_idx = np.argmin(np.abs(fr - f2))
    
                    spec_sum = sum(10 ** (spec[low_limit_idx:high_limit_idx] / 10))
                    Ltot = 10 * np.log10(spec_sum)
    
                    delta_ft = fr[1] - fr[0]
    
            # single tone in a critical band
            else:
                # tone SPL
                Lt = _peak_level(fr, spec, ind_p)
    
                # total SPL in the critical band
                f1, f2 = _critical_band(fr[ind_p])
                low_limit_idx = np.argmin(np.abs(fr - f1))
                high_limit_idx = np.argmin(np.abs(fr - f2))
    
                spec_sum = sum(10 ** (spec[low_limit_idx:high_limit_idx] / 10))
                Ltot = 10 * np.log10(spec_sum)
    
                delta_ft = fr[1] - fr[0]
    
            # SPL of the masking noise
            delta_fc = f2 - f1
            delta_ftot = frs[high_limit_idx] - frs[low_limit_idx]
            Ln = 10 * np.log10(10 ** (Ltot / 10) - 10 ** (Lt / 10)) + 10 * np.log10(
                delta_fc / (delta_ftot - delta_ft)
            )
    
            # Tone-to-noise ratio
            f = fr[ind_p]
            delta_t = Lt - Ln
            if delta_t > 0:
                if nseg > 1:
                    tones_freqs[i] = np.append(tones_freqs[i], f)
                elif nseg == 1:
                    tones_freqs = np.append(tones_freqs, f)
                tnr = np.append(tnr, delta_t)
    
                # Prominence criteria
                if f >= 89.1 and f < 1000:
                    if delta_t >= 8 + 8.33 * np.log10(1000 / f):
                        if nseg > 1:
                            prominence[i].append(True)                        
                        elif nseg == 1:
                            prominence.append(True)
                    else:
                        if nseg > 1:
                            prominence[i].append(False)
                        
                        elif nseg == 1:
                            prominence.append(False)
                elif f >= 1000 and f <= 11200:
                    if delta_t >= 8:
                        if nseg > 1:
                            prominence[i].append(True)
                        
                        elif nseg == 1:
                            prominence.append(True)
                    else:
                        if nseg > 1:
                            prominence[i].append(False)
                        
                        elif nseg == 1:
                            prominence.append(False)
    
            # suppression from the list of tones
            sup = np.where(peaks == ind_p)[0]
            peaks = np.delete(peaks, sup)
            nb_tones -= 1
    

        if nseg > 1:
            if sum(np.power(10, (tnr[prominence[i]] / 10))) != 0:
                t_tnr[i] = 10 * np.log10(sum(np.power(10, (tnr[prominence[i]] / 10))))
            else:
                t_tnr[i] =  0
            TNR[i] = np.append(TNR[i], tnr)               
                    
                    
        elif nseg == 1:
            if sum(np.power(10, (tnr[prominence] / 10))) != 0:
                t_tnr = np.append(t_tnr,10 * np.log10(sum(np.power(10, (tnr[prominence] / 10)))))
            else:
                t_tnr =  0
            TNR = np.append(TNR, tnr)
        
    tones_freqs = np.asarray(tones_freqs, dtype=object)
    prominence = np.asarray(prominence, dtype=object)

    
    return tones_freqs, TNR , prominence, t_tnr



# -*- coding: utf-8 -*-

# Standard library imports
import numpy as np

# Local application imports
from mosqito.sound_level_meter.noct_spectrum._filter_bandwidth import _filter_bandwidth
from mosqito.sound_level_meter.noct_spectrum._n_oct_time_filter import (
    _n_oct_time_filter,
)
from mosqito.sound_level_meter.noct_spectrum._center_freq import _center_freq


def noct_spectrum(sig, fs, fmin, fmax, n=3, G=10, fr=1000):
    """Compute nth-octave band spectrum

    Calculate the rms level of the signal "sig" sampled at freqency "fs"
    for each third octave band between "fc_min" and "fc_max".

    Parameters
    ----------
    sig : numpy.ndarray
        A time signal array with size (nperseg, nseg).
    fs : float
        Sampling frequency [Hz]
    fmin : float
        Min frequency band [Hz]
    fmax : float
        Max frequency band [Hz]
    n : int
        Number of bands pr octave
    G : int
        System for specifying the exact geometric mean frequencies.
        Can be base 2 or base 10
    fr : int
        Reference frequency. Shall be set to 1 kHz for audible frequency
        range, to 1 Hz for infrasonic range (f < 20 Hz) and to 1 MHz for
        ultrasonic range (f > 31.5 kHz)

    Outputs
    -------
    spec : numpy.ndarray
        The third octave band spectrum of signal sig with size (nfreq, nseg)
    fpref : numpy.ndarray
        Corresponding preferred third octave band center frequencies
    """

    # 1-dimensional array to 2-dimensional array with size (nperseg, 1)
    if sig.ndim == 1:
        sig = sig[:, np.newaxis]

    # Get filters center frequencies
    fc_vec, fpref = _center_freq(fmin=fmin, fmax=fmax, n=n, G=G, fr=fr)

    # Compute the filters bandwidth
    alpha_vec, _, _ = _filter_bandwidth(fc_vec, n=n)

    # Calculation of the rms level of the signal in each band
    spec = []
    for fc, alpha in zip(fc_vec, alpha_vec):
        spec.append(_n_oct_time_filter(sig, fs, fc, alpha))

    return np.array(spec), fpref



# -*- coding: utf-8 -*-

# Local functions imports
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._tnr_main_calc import _tnr_main_calc


def tnr_ecma_st(signal, fs, prominence=True):
    """Computation of tone-to-noise ration according to ECMA-74, annex D.9
    for a stationary signal.
        The T-TNR value is calculated according to ECMA-TR/108

    Parameters
    ----------
    signal :numpy.array
        A stationary signal in [Pa].
    fs : integer
        Sampling frequency.
    prominence : boolean
        If True, the algorithm only returns the prominent tones, if False it returns all tones detected.
        Default is True.

    Output
    ------
    t_tnr : array of float
        global TNR value, along time if is_stationary = False
    tnr : array of float
        TNR values for each detected tone
    promi : array of bool
        prominence criterion for each detected tone
    tones_freqs : array of float
        frequency of the detected tones
    """
    
    # Compute db spectrum
    spectrum_db, freq_axis = spectrum(signal, fs, db=True)
        
    # Compute tnr values
    tones_freqs, tnr, prom, t_tnr = _tnr_main_calc(spectrum_db, freq_axis)
    prom = prom.astype(bool)
    
    if prominence == False:
        return t_tnr, tnr, prom, tones_freqs
    else:
        return t_tnr, tnr[prom], prom[prom], tones_freqs[prom]
    



# -*- coding: utf-8 -*-

# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError(
        "In order to perform the tests you need the 'pytest' package."
    )
# External import
import numpy as np

# Local application imports
from mosqito.utils import load
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sq_metrics import pr_ecma_freq

@pytest.mark.pr_freq  # to skip or run PR test
def test_pr_ecma_freq():
    """Test function for the prominence ratio calculation of an audio signal

    Validation function for the "pr_ecma_freq" function with complex spectrum array
    as input. The input signal was generated using audacity, and then the spectrum is computed using mosqito.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    # Test signal as input for prominence ratio calculation
    # signals generated using audacity : white noise + tones at 200 and 2000 Hz


    signal =         {
            "is_stationary": True,
            "data_file": "tests/input/white_noise_442_1768_Hz_stationary.wav"
        }


    # Load signal
    audio, fs = load(signal["data_file"], wav_calib=0.01)
    # convert to frequency domain
    spec, freqs = spectrum(audio, fs, window='hanning', db=False)

    # Compute tone-to-noise ratio
    # 1D input
    # Compute tone-to-noise ratio
    t_pr, pr, promi, freq = pr_ecma_freq(spec, freqs=freqs, prominence=False )

    # 2D spectrum, 1D freq axis
    spec = np.tile(spec, (4,1)).T
    
    t_pr, pr, promi, freq = pr_ecma_freq(spec, freqs=freqs, prominence=False )

    # 2D spectrum, 2D freq axis
    freqs = np.tile(freqs,(4,1)).T
    t_pr, pr, promi, freq = pr_ecma_freq(spec, freqs=freqs, prominence=False )
        
if __name__ == "__main__":
    test_pr_ecma_freq()




# -*- coding: utf-8 -*-

# Local imports
from mosqito.sq_metrics import loudness_zwst_perseg
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_from_loudness import (
    sharpness_din_from_loudness,
)

# Optional package import
try:
    from SciDataTool import DataTime, DataLinspace, DataFreq, Norm_func
except ImportError:
    DataTime = None
    DataLinspace = None
    DataFreq = None


def sharpness_din_perseg(
    signal,
    fs=None,
    weighting="din",
    nperseg=4096,
    noverlap=None,
    field_type="free",
    is_sdt_output=False,
):
    """Acoustic sharpness calculation according to different methods
        (Aures, Von Bismarck, DIN 45692, Fastl) from a stationary signal.

    Parameters:
    ----------
    signal: numpy.array or DataTime object
        A time signal in [Pa].
    fs : float, optional
        Sampling frequency, can be omitted if the input is a DataTime
        object. Default to None
    weighting : string
        To specify the weighting function used for the
        sharpness computation.'din' by default,'aures', 'bismarck','fastl'
    nperseg: int, optional
        Length of each segment. Defaults to 4096.
    noverlap: int, optional
        Number of points to overlap between segments.
        If None, noverlap = nperseg / 2. Defaults to None.
    field_type : str
        Type of soundfield corresponding to spec_third ("free" by
        default or "diffuse").
    is_sdt_output : Bool, optional
        If True, the outputs are returned as SciDataTool objects.
        Default to False

    Outputs
    ------
    S : ndarray or DataTime object
        Sharpness value, size(nseg).

    """

    # Manage input type
    if DataTime is not None and isinstance(signal, DataTime):
        time = signal.get_along("time")["time"]
        fs = 1 / (time[1] - time[0])
        signal = signal.get_along("time")[signal.symbol]

    # Compute loudness
    N, N_specific, _, time_axis = loudness_zwst_perseg(
        signal, fs, nperseg=nperseg, noverlap=noverlap, field_type=field_type
    )

    # Compute sharpness from loudness
    S = sharpness_din_from_loudness(N, N_specific, weighting=weighting)

    # Manage SciDataTool output type
    if is_sdt_output:
        if DataLinspace is None:
            raise RuntimeError(
                "In order to handle Data objects you need the 'SciDataTool' package."
            )
        else:
            time = DataLinspace(
                name="time",
                unit="s",
                initial=time_axis[0],
                final=time_axis[-1],
                number=len(time_axis),
                include_endpoint=True,
            )
            S = DataTime(
                name="Sharpness (DIN 45692)",
                symbol="S_{DIN}",
                axes=[time],
                values=S,
                unit="acum",
            )

    return S, time_axis



# -*- coding: utf-8 -*-

from numpy import ones, log10, power

from mosqito.sq_metrics.speech_intelligibility.sii_ansi._band_procedure_data import _get_critical_band_data, _get_equal_critical_band_data, _get_octave_band_data, _get_third_octave_band_data
from mosqito.sq_metrics.speech_intelligibility.sii_ansi._speech_data import _get_critical_band_speech_data, _get_equal_critical_band_speech_data, _get_octave_band_speech_data, _get_third_octave_band_speech_data
from mosqito.sq_metrics.speech_intelligibility.sii_ansi._main_sii import _main_sii


def sii_ansi_level(noise_level, method, speech_level, threshold=None):
    """Calculate speech intelligibility index

    This function computes SII values for an overall noise level in dB according to ANSI S3.5 standard.
    This value is used to create a uniform noise spectrum, with the same deduced level on each frequency band.

    Parameters
    ----------
    noise_level : float
        Overall noise level in [dB ref. 2e-5 Pa]. 
    method: {"critical", "equally_critical", "third_octave", "octave"}
        Type of frequency band to be used for the calculation. See § 3.4 of the standard.
    speech_level : {'normal', 'raised', 'loud', 'shout'}
        Speech level to assess, the corresponding speech spectrum defined in the standard is used for calculation.
    threshold : array_like or 'zwicker'
        Threshold of hearing [dB ref. 2e-5 Pa] with same size as the chosen method frequency axis, or 'zwicker' to use the standard threshold.
        Default to None sets the threshold to zeros on each frequency band.
                
    Returns
    -------
    sii: numpy.ndarray
        Overall SII value.
    specific_sii: numpy.ndarray
        Specific SII values along the frequency axis.
    freq_axis: numpy.ndarray
        Frequency axis corresponding to the chosen method.
        
    Examples
    --------
    .. plot::
       :include-source:
       
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from mosqito.sq_metrics.speech_intelligibility import sii_level
        >>> fs=48000
        >>> d=0.2
        >>> dB=90
        >>> time = np.arange(0, d, 1/fs)
        >>> f = 50
        >>> stimulus = np.sin(2 * np.pi * f * time) * np.sin(np.pi * f * time) + np.sin(10 * np.pi * f * time) + np.sin(100 * np.pi * f * time)
        >>> rms = np.sqrt(np.mean(np.power(stimulus, 2)))
        >>> ampl = 0.00002 * np.power(10, dB / 20) / rms
        >>> stimulus = stimulus * ampl
        >>> speech_level = 'raised'
        >>> SII, SII_spec, freq_axis = sii_level(60, method='critical', speech_level=speech_level, threshold='zwicker')
        >>> plt.plot(freq_axis, SII_spec)
        >>> plt.xlabel("Frequency [Hz]")
        >>> plt.ylabel("Specific value ")
        >>> plt.title("Speech Intelligibility Index = " + f"{SII:.2f} \n Speech level: " + speech_level)   

    """
    
    if (method!='critical') & (method!='equally_critical') & (method!='third_octave') & (method!='octave'):
        raise ValueError('Method should be within {"critical", "equally_critical", "third_octave", "octave"}.')
        
    if (speech_level!='normal') & (speech_level!='raised') & (speech_level!='loud') & (speech_level!='shout'):
        raise ValueError('Speech level should be within {"normal", "raised", "loud", "shout"} to use the corresponding standard data.')
    
    # Get standard speech spectrum
    if method == 'critical':
        speech_spectrum, speech_level = _get_critical_band_speech_data(speech_level)
    elif method == 'equally_critical':
        speech_spectrum, speech_level = _get_equal_critical_band_speech_data(speech_level)
    elif method == 'third_octave':
        speech_spectrum, speech_level = _get_third_octave_band_speech_data(speech_level)
    elif method == 'octave':
        speech_spectrum, speech_level = _get_octave_band_speech_data(speech_level)
    nbands = len(speech_spectrum)
    
    # Create noise spectrum as uniform spectrum from overall level
    band_level = 10 * log10(power(10, noise_level/10)/nbands)
    noise_spectrum = ones((nbands)) * band_level
        
    # Compute SII
    SII, SII_specific, freq_axis = _main_sii(method, speech_spectrum, noise_spectrum, threshold)    
    
    return SII, SII_specific, freq_axis




# -*- coding: utf-8 -*-

# Local imports
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sq_metrics.tonality.prominence_ratio_ecma._pr_main_calc import _pr_main_calc


def pr_ecma_st(signal, fs, prominence=True):
    """Computation of prominence ratio according to ECMA-74, annex D.10
    for a stationary signal.
        The T-PR value is calculated according to ECMA-TR/108

    Parameters
    ----------
    signal :numpy.array
        A stationary signal in [Pa].
    fs : integer
        Sampling frequency.
    prominence : boolean
        If True, the algorithm only returns the prominent tones, if False it returns all tones detected.
        Default is True.

    Output
    ------
    t_pr : array of float
        Global PR value along time.
    pr : array of float
        PR values for each detected tone.
    promi : array of bool
        Prominence criterion for each detected tone.
    tones_freqs : array of float
        Frequency list of the detected tones.
    """
    
    # Compute db spectrum
    spectrum_db, freq_axis = spectrum(signal, fs, db=True)
                  
    # Compute PR values
    tones_freqs, pr, prom, t_pr = _pr_main_calc(spectrum_db, freq_axis)
    prom = prom.astype(bool)
    
    if prominence == False:
        return t_pr, pr, prom, tones_freqs
    else:
        return t_pr, pr[prom], prom[prom], tones_freqs[prom]



# -*- coding: utf-8 -*-
"""
@author: Daniel Jiménez-Caminero Costa
"""
import numpy as np

# Project Imports
from mosqito.sq_metrics.loudness.loudness_ecma._rectified_band_pass_signals import (
    _rectified_band_pass_signals,
)
from mosqito.sq_metrics.loudness.loudness_ecma._nonlinearity import _nonlinearity

# Data import
# Threshold in quiet
from mosqito.sq_metrics.loudness.loudness_ecma._loudness_ecma_data import ltq_z


def loudness_ecma(signal, sb=2048, sh=1024):
    """Calculation of the specific and total loudness according to ECMA-418-2 section 5

    Parameters
    ----------
    signal: numpy.array
        time signal values in 'Pa'. The sampling frequency of the signal must be 48000 Hz.
    sb: int or list of int
        block size.
    sh: int or list of int
        Hop size.

    Returns
    -------
    n_specific: list of numpy.array
        Specific Loudness [sone_HMS per Bark]. Each of the 53 element of the list corresponds to the time-dependant
        specific loudness for a given bark band. Can be a ragged array if a different sb/sh are used for each band.

    bark_axis: numpy.array
        Bark axis

    """

    # TODO: return time axis (list or list of list)

    # Computaton of rectified band-pass signals
    # (section 5.1.2 to 5.1.5 of the standard)
    block_array_rect = _rectified_band_pass_signals(signal, sb, sh)

    # # sb and sh for Tonality
    # z = np.linspace(0.5, 26.5, num=53, endpoint=True)
    # sb = np.ones(53, dtype="int")
    # sh = np.ones(53, dtype="int")
    # sb[z <= 1.5] = 8192
    # sh[z <= 1.5] = 2048
    # sb[np.all([z >= 2, z <= 8], axis=0)] = 4096
    # sh[np.all([z >= 2, z <= 8], axis=0)] = 1024
    # sb[np.all([z >= 8.5, z <= 12.5], axis=0)] = 2048
    # sh[np.all([z >= 8.5, z <= 12.5], axis=0)] = 512
    # sb[z >= 13] = 1024
    # sh[z >= 13] = 256

    n_specific = []
    for band_number in range(53):
        # ROOT-MEAN-SQUARE (section 5.1.6)
        # After the segmentation of the signal into blocks, root-mean square values of each block are calculated
        # according to Formula 17.
        rms_block_value = np.sqrt(
            2 * np.mean(np.array(block_array_rect[band_number]) ** 2, axis=1)
        )

        # NON-LINEARITY (section 5.1.7)
        # This section covers the other part of the calculations needed to consider the non-linear transformation
        # of sound pressure to specific loudness that does the the auditory system. After this point, the
        # computation is done equally to every block in which we have divided our signal.
        a_prime = _nonlinearity(rms_block_value)

        # SPECIFIC LOUDNESS CONSIDERING THE THRESHOLD IN QUIET (section 5.1.8)
        # The next calculation helps us obtain the result for the specific loudness - specific loudness with
        # consideration of the lower threshold of hearing.
        a_prime[a_prime < ltq_z[band_number]] = ltq_z[band_number]
        N_prime = a_prime - ltq_z[band_number]
        n_specific.append(N_prime)

    bark_axis = np.linspace(0.5, 26.5, num=53, endpoint=True)
    return n_specific, bark_axis



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 16 20:23:01 2020

@author: Salomé
"""
# Standard library imports
import numpy as np

# Mosqito functions import
from mosqito.sq_metrics.tonality.tone_to_noise_ecma._critical_band import _critical_band


def _find_highest_tone(freqs, spec_db, index, nb_tones, ind):
    """

    Method to find the two highest tones in a given spectrum from a given index
    according to their critical band

    Parameters
    ----------
    freqs : numpy.array
        frequency axis
    spec_db : numpy.array
        signal spectrum in dB
    index : numpy.array
        list of candidate tones index
    index : numpy.array
        list of candidate tones index
    nb_tones : integer
        number of candidate tones non examinated

    Returns
    -------
    ind_p : integer
        index of the highest tone in the critical band
    ind_p : integer
        index of the second highest tone in the critical band
    index : numpy.array
        list of candidate tones index updated
    nb_tones : integer
        number of candidate tones non examinated updated
    """

    f = freqs[ind]
    # critical band centered on f
    f1, f2 = _critical_band(f)
    low_limit_idx = np.argmin(np.abs(freqs - f1))
    high_limit_idx = np.argmin(np.abs(freqs - f2))

    # Other tones in the critical band centered on f tones
    multiple_idx = index[index > low_limit_idx]
    multiple_idx = multiple_idx[multiple_idx < high_limit_idx]

    if len(multiple_idx) > 1:
        sort_spec = np.argsort(-1 * spec_db[multiple_idx])

        # highest tones in the critical band
        ind_p = multiple_idx[sort_spec[0]]
        ind_s = multiple_idx[sort_spec[1]]

        # suppression of the lower values
        for s in sort_spec[2:]:
            sup = np.where(index == multiple_idx[s])[0]
            index = np.delete(index, sup)
            nb_tones -= 1

        if ind_p != ind:
            # screening to find the highest value in the critical band centered on fp
            ind_p, ind_s, index, nb_tones = _find_highest_tone(
                freqs, spec_db, index, nb_tones, ind_p
            )

    else:
        ind_p = ind
        ind_s = None

    return ind_p, ind_s, index, nb_tones



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 10:26:48 2020

@author: wantysal
"""
# Standard library import
import numpy as np


def _peak_level(freqs, spec, peak_index):
    """
        Correct the level of the peak present at peak_index

        This kind of error in peak levels is due to spectrum resolution
        The algorithm sums up the levels which are strictly decreasing
        from the peak with a level difference under 10 dB.

    Parameters
    ----------
    freqs : numpy.array
        frequency axis
    spec : numpy.array
        SPL spectrum in dB
    peak_index : integer
        index where to find the peak in the spectrum

    Returns
    -------
    Lt : float
        corrected SPL at peak_index

    """

    # initial level
    Li = spec[peak_index]
    # modified level
    L = spec[peak_index]

    # Screen the right points of the peak
    temp = peak_index + 1
    
    if temp != len(spec):
        Ltemp = Li
        # As long as the level decreases,
        while Ltemp - np.abs(spec[temp]) > 0:
            # if the level of the point is close enough of the peak point,
            if Li - spec[temp] < 10:
                Ltemp = spec[temp]
                # its level is summed up with the peak's one
                L = 10 * np.log10(10 ** (L / 10) + 10 ** (spec[temp] / 10))
                
                temp += 1
                if temp == len(spec):
                    temp -= 1
                    Ltemp = -1
            else:
                Ltemp = -1

    # Screen the left points of the peak
    temp = peak_index - 1
    
    if temp != -1:
        Ltemp = Li
        # As long as the level decreases,
        while Ltemp - np.abs(spec[temp]) > 0:
            # if the level of the point is close enough of the peak point,
            if Li - spec[temp] < 10:
                Ltemp = spec[temp]
                # its level is summed up with the peak's one
                L = 10 * np.log10(10 ** (L / 10) + 10 ** (spec[temp] / 10))
    
                temp -= 1
                if temp <0:
                    temp += 1
                    Ltemp = -1
            else:
                Ltemp = -1

    return L



# -*- coding: utf-8 -*-

# External import
import numpy as np

# Local imports
from mosqito.sq_metrics.tonality.prominence_ratio_ecma._pr_main_calc import _pr_main_calc
from mosqito.utils.conversion import amp2db


def pr_ecma_freq(spectrum, freqs, prominence=True):
    """Computation of prominence ratio according to ECMA-74, annex D.10
    for an amplitude or complex frequency spectrum.
        The T-PR value is calculated according to ECMA-TR/108

    Parameters
    ----------
    spectrum :numpy.array
        An amplitude or complex frequency spectrum [nperseg x nseg].
    freqs : np.array
        Frequency axis [nperseg x nseg] or [nperseg]. 
    prominence : boolean
        If True, the algorithm only returns the prominent tones, if False it returns all tones detected.
        Default is True.

    Output
    ------
    t_pr : array of float
        Global PR value.
    pr : array of float
        PR values for each detected tone.
    promi : array of bool
        Prominence criterion for each detected tone.    
    tones_freqs : array of float
        Frequency list of the detected tones.
    """
    

    if len(spectrum) != len(freqs) :
        raise ValueError('Input spectrum and frequency axis must have the same size')

    # Compute spectrum dB values
    spectrum_db = amp2db(np.abs(spectrum), ref=2e-5)
                  
    # Compute PR values
    tones_freqs, pr, prom, t_pr = _pr_main_calc(spectrum_db, freqs)
 
    if prominence == False:
        return t_pr, pr, prom, tones_freqs  
    else:
        return t_pr, pr[prom], prom[prom], tones_freqs[prom]



from mosqito.utils.isoclose import isoclose
from mosqito.utils.load import load
from mosqito.utils.sine_wave_generator import sine_wave_generator
from mosqito.utils.time_segmentation import time_segmentation



# -*- coding: utf-8 -*-

# Third party imports
import numpy as np

# Local application imports
from mosqito.sound_level_meter.noct_spectrum.noct_synthesis import noct_synthesis
from mosqito.sq_metrics.loudness.loudness_zwst._main_loudness import _main_loudness
from mosqito.sq_metrics.loudness.loudness_zwst._calc_slopes import _calc_slopes
from mosqito.utils.conversion import amp2db


def loudness_zwst_freq(spectrum, freqs, field_type="free"):
    """Zwicker-loudness calculation for stationary signals

    Calculates the acoustic loudness according to Zwicker method for
    stationary signals.
    Normatice reference:
        ISO 532:1975 (method B)
        DIN 45631:1991
        ISO 532-1:2017 (method 1)
    The code is based on BASIC program published in "Program for
    calculating loudness according to DIN 45631 (ISO 532B)", E.Zwicker
    and H.Fastl, J.A.S.J (E) 12, 1 (1991).
    Note that due to normative continuity, as defined in the
    preceeding standards, the method is in accordance with
    ISO 226:1987 equal loudness contours (instead of ISO 226:2003)

    Parameters
    ----------
    spectrum : numpy.array
        A RMS frequency spectrum, size (Nfreq, Ntime)
    freqs : list
        List of the corresponding frequencies, size (Nfreq,) or (Nfreq, Ntime)
    field_type : str
        Type of soundfield corresponding to spec_third ("free" by
        default or "diffuse")

    Outputs
    -------
    N : float or numpy.array
        Calculated loudness [sones], size (Ntime,).
    N_specific : numpy.ndarray
        Specific loudness [sones/bark], size (Nbark, Ntime).
    bark_axis : numpy.array
        Frequency axis in bark, size (Nbark,).
    """
    if len(spectrum) != len(freqs):
        raise ValueError('Input spectrum and frequency axis must have the same shape')

    # Compute third octave band spectrum
    spec_third, _ = noct_synthesis(spectrum, freqs, fmin=24, fmax=12600)

    # Compute dB values
    spec_third = amp2db(spec_third, ref=2e-5)

    # Compute main loudness
    Nm = _main_loudness(spec_third, field_type)
    #
    # Computation of specific loudness pattern and integration of overall
    # loudness by attaching slopes towards higher frequencies
    N, N_specific = _calc_slopes(Nm)

    # Define Bark axis
    bark_axis = np.linspace(0.1, 24, int(24 / 0.1))

    return N, N_specific, bark_axis



# -*- coding: utf-8 -*-


# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError(
        "In order to perform the tests you need the 'pytest' package."
    )

import numpy as np
# Local application imports
from mosqito.utils import load
from mosqito.sq_metrics import tnr_ecma_st


@pytest.mark.tnr_st  # to skip or run PR test
def test_tnr_ecma_st():
    """Test function for the prominence ratio calculation of an audio signal

    Validation function for the "tnr_ecma_st" function with stationary signal array
    as input. The input signal was generated using audacity.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    # Test signal as input for tone to noise ratio calculation
    # signals generated using audacity : white noise + tones at 442 and 1768 Hz

    signal =  {
            "tones freq": [442, 1768],
            "data_file": "tests/input/white_noise_442_1768_Hz_stationary.wav"
        }

    # Load signal
    audio, fs = load(signal["data_file"], wav_calib=0.01)
    # Compute tone-to-noise ratio
    t_tnr, tnr, prom, freq = tnr_ecma_st(audio, fs, prominence=True)
    np.testing.assert_almost_equal(t_tnr, 32.7142766)
    np.testing.assert_almost_equal(freq.astype(np.int32), [442, 1768])
    assert np.count_nonzero(prom == True) == 2
        
if __name__ == "__main__":
    test_tnr_ecma_st()





# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 10:11:27 2020

@author: pc
"""

import numpy as np


def _ear_filter_coeff(bark_axis):
    """Zwicker coefficient for the transmission between outer and inner ear

    See E. Zwicker, H. Fastl: Psychoacoustics. Springer,Berlin, Heidelberg, 1990.
    The coefficients are linearly interpolated from the values given in figure 8.18

    Parameter
    ---------
    bark_axis: numpy.array
        frequency axis in Bark

    Output
    ------
    a0tab: numpy.array
        coefficients a0 along the given bark axis
    """

    xp = [
        0,
        10,
        12,
        13,
        14,
        15,
        16,
        16.5,
        17,
        18,
        18.5,
        19,
        20,
        21,
        21.5,
        22,
        22.5,
        23,
        23.5,
        24,
        25,
        26,
    ]

    yp = [
        0,
        0,
        1.15,
        2.31,
        3.85,
        5.62,
        6.92,
        7.38,
        6.92,
        4.23,
        2.31,
        0,
        -1.43,
        -2.59,
        -3.57,
        -5.19,
        -7.41,
        -11.3,
        -20,
        -40,
        -130,
        -999,
    ]

    a0tab = np.interp(bark_axis, xp, yp)

    return a0tab



from mosqito.sound_level_meter.noct_spectrum.noct_spectrum import noct_spectrum
from mosqito.sound_level_meter.noct_spectrum.noct_synthesis import noct_synthesis
from mosqito.sound_level_meter.spectrum import spectrum
from mosqito.sound_level_meter.freq_band_synthesis import freq_band_synthesis



# -*- coding: utf-8 -*-


# Local imports
from mosqito.sq_metrics import loudness_zwst_freq
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_from_loudness import sharpness_din_from_loudness


def sharpness_din_freq(spectrum, freqs, weighting="din", field_type="free"):
    """Acoustic sharpness calculation according to different methods
      (Aures, Von Bismarck, DIN 45692, Fastl) from a complex spectrum.

    Parameters:
    ----------
    signal: numpy.array
        A RMS spectrum.
    freqs: integer
        Frequency axis.
    method : string
        To specify the Loudness computation method
    weighting : string
        To specify the weighting function used for the
        sharpness computation.'din' by default,'aures', 'bismarck','fastl'
    field_type : str
        Type of soundfield corresponding to spec_third ("free" by
        default or "diffuse").

    Outputs
    ------
    S : float
        sharpness value
    time_axis: numpy.array
        The time axis array, size (Ntime,) or None

    """
    if len(spectrum) != len(freqs):
        raise ValueError(
            'Input spectrum and frequency axis must have the same shape')

    # Compute loudness
    N, N_specific, _ = loudness_zwst_freq(
        spectrum, freqs, field_type=field_type)

    if len(spectrum.shape) > 1:
        raise ValueError(
            "With a 2D spectrum use 'sharpness_din_perseg' calculation.")

    # Compute sharpness from loudness
    S = sharpness_din_from_loudness(N, N_specific, weighting=weighting)

    return S



# -*- coding: utf-8 -*-

# Standard imports
import numpy as np

# Local imports
from mosqito.sq_metrics.roughness.roughness_dw._roughness_dw_main_calc import _roughness_dw_main_calc
from mosqito.sq_metrics.roughness.roughness_dw._gzi_weighting import _gzi_weighting
from mosqito.sq_metrics.roughness.roughness_dw._H_weighting import _H_weighting


def roughness_dw_freq(spectrum, freqs):
    """Roughness calculation of a spectrum sampled at 48kHz.

    The code is based on the algorithm described in "Psychoacoustical roughness:
    implementation of an optimized model" by Daniel and Weber in 1997.
    The roughness model consists of a parallel processing structure that is made up
    of successive stages and calculates intermediate specific roughnesses R_spec,
    which are summed up to determine the total roughness R.

    Parameters
    ----------
    spectrum :numpy.array
        An amplitude or complex frequency spectrum [nperseg x nseg].
    freqs : np.array
        Frequency axis [nperseg] if identical for all the blocks, [nperseg x nseg] if not.

    Outputs
    -------
    R : numpy.array
        Roughness in [asper], dim [nseg].
    R_spec : numpy.array
        Specific roughness over bark axis, dim [47 bark x nseg].
    bark_axis : numpy.array
        Frequency axis in [bark], dim [nseg].

    """

    # Check input size coherence
    if len(spectrum) != len(freqs):
        raise ValueError(
            'Input spectrum and frequency axis must have the same size !')

    if spectrum.any() < 0:
        raise ValueError(
            'Input must be an amplitude spectrum (use np.abs() or complex spectrum).')

    # 1D spectrum
    if len(spectrum.shape) == 1:
        nperseg = len(spectrum)
        nseg = 1
        fs = int(2 * nperseg * np.mean(freqs[1:] - freqs[:-1]))

    # 2D spectrum
    elif len(spectrum.shape) > 1:
        nperseg = spectrum.shape[0]
        nseg = spectrum.shape[1]
        # one frequency axis per block
        if len(freqs.shape) > 1:
            fs = int(2 * nperseg * np.mean(freqs[0, 1:] - freqs[0, :-1]))
        # one frequency axis for all the blocks
        elif len(freqs.shape) == 1:
            fs = int(2 * nperseg * np.mean(freqs[1:] - freqs[:-1]))
            freqs = np.tile(freqs, (nseg, 1)).T

    # Initialization of the weighting functions H and g
    hWeight = _H_weighting(2*nperseg, fs)
    # Aures modulation depth weighting function
    gzi = _gzi_weighting(np.arange(1, 48, 1) / 2)

    R = np.zeros((nseg))
    R_spec = np.zeros((47, nseg))

    if len(spectrum.shape) > 1:
        for i in range(nseg):
            R[i], R_spec[:, i], bark_axis = _roughness_dw_main_calc(
                spectrum[:, i], freqs[:, i], fs, gzi, hWeight)
    else:
        R, R_spec, bark_axis = _roughness_dw_main_calc(
            spectrum, freqs, fs, gzi, hWeight)

    return R, R_spec, bark_axis



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 13 15:17:27 2020

@author: wantysal
"""
# Standard library import
import numpy as np
from scipy.io.wavfile import write


def signal_test(fc, fmod, mdepth, fs, d, dB):
    """Creation of stationary amplitude modulated signals for the roughness
    validation procedure (signal created according to equation 1 in
    "Psychoacoustical roughness:implementation of an optimized model"
    by Daniel and Weber in 1997.

    Parameters
    ----------
    fc: integer
        carrier frequency
    fmod: integer
        modulation frequency
    mdepth: float
        modulation depth
    fs: integer
        sampling frequency
    d: float
        signal duration [s]
    dB: integer
        SPL dB level of the carrier signal
    """

    # time axis definition
    dt = 1 / fs
    time = np.arange(0, d, dt)

    signal = (
        0.5
        * (1 + mdepth * (np.sin(2 * np.pi * fmod * time)))
        * np.sin(2 * np.pi * fc * time)
    )
    rms = np.sqrt(np.mean(np.power(signal, 2)))
    ampl = 0.00002 * np.power(10, dB / 20) / rms
    signal = signal * ampl

    return signal, time


def wav_test(fc, fmod, mdepth, fs, d, dB, folder):
    """Creation of .wav file of stationary amplitude modulated signals for the roughness
    validation procedure (signal created according to equation 1 in
    "Psychoacoustical roughness:implementation of an optimized model"
    by Daniel and Weber in 1997.

    Parameters
    ----------
    fc: integer
        carrier frequency
    fmod: integer or numpy.array
        modulation frequency
    mdepth: float
        modulation depth
    fs: integer
        sampling frequency
    d: float
        signal duration [s]
    dB: integer
        SPL dB level of the carrier signal
    folder: string
        path of the folder where to store the file
    """

    values = signal_test(fc, fmod, mdepth, fs, d, dB)
    values = values / (2 * 2 ** 0.5)
    values = values.astype(np.int16)
    write(
        folder + "/Test_signal_fc" +
        str(fc) + "_fmod" + str(fmod) + ".wav", fs, values
    )



# -*- coding: utf-8 -*-

# Local application imports
from mosqito.utils import time_segmentation
from mosqito.sq_metrics.loudness.loudness_zwst.loudness_zwst import loudness_zwst

# Optional package import
try:
    from SciDataTool import DataTime, DataLinspace, DataFreq, Norm_func
except ImportError:
    DataTime = None
    DataLinspace = None
    DataFreq = None


def loudness_zwst_perseg(
    signal, fs=None, nperseg=4096, noverlap=None, field_type="free", is_sdt_output=False
):
    """Zwicker-loudness calculation for stationary signals

    Calculates the acoustic loudness according to Zwicker method for
    stationary signals per signal segment.
    Normatice reference:
        ISO 532:1975 (method B)
        DIN 45631:1991
        ISO 532-1:2017 (method 1)
    The code is based on BASIC program published in "Program for
    calculating loudness according to DIN 45631 (ISO 532B)", E.Zwicker
    and H.Fastl, J.A.S.J (E) 12, 1 (1991).
    Note that due to normative continuity, as defined in the
    preceeding standards, the method is in accordance with
    ISO 226:1987 equal loudness contours (instead of ISO 226:2003)

    Parameters
    ----------
    signal : numpy.array or DataTime object
        Time signal values [Pa].
    fs : float, optional
        Sampling frequency, can be omitted if the input is a DataTime
        object. Default to None
    nperseg: int, optional
        Length of each segment. Defaults to 4096.
    noverlap: int, optional
        Number of points to overlap between segments.
        If None, noverlap = nperseg / 2. Defaults to None.
    field_type : str
        Type of soundfield corresponding to spec_third ("free" by
        default or "diffuse").
    is_sdt_output : Bool, optional
        If True, the outputs are returned as SciDataTool objects.
        Default to False

    Outputs
    -------
    N : numpy.array or DataTime object
        The overall loudness array [sones], size (Ntime,).
    N_specific : numpy.ndarray or DataFreq object
        The specific loudness array [sones/bark], size (Nbark, Ntime).
    bark_axis: numpy.array
        The Bark axis array, size (Nbark,).
    time_axis: numpy.array
        The time axis array, size (Ntime,) or None.

    """

    # Manage input type
    if DataTime is not None and isinstance(signal, DataTime):
        time = signal.get_along("time")["time"]
        fs = 1 / (time[1] - time[0])
        signal = signal.get_along("time")[signal.symbol]

    # Time signal segmentation
    signal, time_axis = time_segmentation(signal, fs, nperseg, noverlap)

    # Compute loudness
    N, N_specific, bark_axis = loudness_zwst(signal, fs, field_type=field_type)

    # Manage SciDataTool output type
    if is_sdt_output:
        if DataLinspace is None:
            raise RuntimeError(
                "In order to handle Data objects you need the 'SciDataTool' package."
            )
        else:
            bark_data = DataLinspace(
                name="Critical band rate",
                unit="Bark",
                initial=0,
                final=24,
                number=int(24 / 0.1),
                include_endpoint=True,
                normalizations={
                    "Hz": Norm_func(function=lambda x: 1960 * (x + 0.53) / (26.28 - x))
                },
            )
            time = DataLinspace(
                name="time",
                unit="s",
                initial=time_axis[0],
                final=time_axis[-1],
                number=len(time_axis),
                include_endpoint=True,
            )
            N_specific = DataFreq(
                name="Specific loudness (Zwicker method for stationnary signal)",
                symbol="N'_{zwst}",
                axes=[bark_data, time],
                values=N_specific,
                unit="sone/Bark",
            )
            N = DataTime(
                name="Loudness (Zwicker method for stationnary signal)",
                symbol="N_{zwst}",
                axes=[time],
                values=N,
                unit="sone",
            )

    return N, N_specific, bark_axis, time_axis



# -*- coding: utf-8 -*-
"""
@date Created on Fri May 22 2020
@author martin_g for Eomys
"""

# Local application import
from mosqito.sq_metrics.loudness.loudness_zwtv._lowpass_intp import (
    _lowpass_intp,
)


def _temporal_weighting(loudness):
    """Temporal weighting of total loudness

    Two first-order low-pass filters (time constants 3,5 ms
    and 70 ms) are applied to the sum of the specific loudness
    values in order to simulate the duration dependent behaviour
    of loudness perception for short impulses.

    Parameters
    ----------
    loudness : numpy.ndarray
        Loudness vs. time

    Outputs
    -------
    loudness : numpy.ndarray
        Filtered loudness
    """
    sample_rate = 2000
    tau = 3.5 * 10 ** -3
    filt_loudness_1 = _lowpass_intp(loudness, tau, sample_rate)
    tau = 70 * 10 ** -3
    filt_loudness_2 = _lowpass_intp(loudness, tau, sample_rate)

    loudness = 0.47 * filt_loudness_1 + 0.53 * filt_loudness_2

    return loudness



from mosqito.sq_metrics.loudness.loudness_ecma.loudness_ecma import loudness_ecma
from mosqito.sq_metrics.loudness.loudness_zwst.loudness_zwst import loudness_zwst
from mosqito.sq_metrics.loudness.loudness_zwst.loudness_zwst_freq import (
    loudness_zwst_freq,
)
from mosqito.sq_metrics.loudness.loudness_zwst.loudness_zwst_perseg import (
    loudness_zwst_perseg,
)

from mosqito.sq_metrics.loudness.loudness_zwtv.loudness_zwtv import loudness_zwtv
from mosqito.sq_metrics.loudness.utils.equal_loudness_contours import (
    equal_loudness_contours,
)

from mosqito.sq_metrics.tonality.prominence_ratio_ecma.pr_ecma_st import pr_ecma_st
from mosqito.sq_metrics.tonality.prominence_ratio_ecma.pr_ecma_tv import pr_ecma_tv
from mosqito.sq_metrics.tonality.prominence_ratio_ecma.pr_ecma_freq import pr_ecma_freq

from mosqito.sq_metrics.tonality.tone_to_noise_ecma.tnr_ecma_st import tnr_ecma_st
from mosqito.sq_metrics.tonality.tone_to_noise_ecma.tnr_ecma_tv import tnr_ecma_tv
from mosqito.sq_metrics.tonality.tone_to_noise_ecma.tnr_ecma_freq import tnr_ecma_freq

from mosqito.sq_metrics.roughness.roughness_dw.roughness_dw import roughness_dw
from mosqito.sq_metrics.roughness.roughness_dw.roughness_dw_freq import (
    roughness_dw_freq,
)

from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_st import sharpness_din_st
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_tv import sharpness_din_tv
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_from_loudness import sharpness_din_from_loudness
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_perseg import sharpness_din_perseg
from mosqito.sq_metrics.sharpness.sharpness_din.sharpness_din_freq import sharpness_din_freq

from mosqito.sq_metrics.loudness.utils.sone_to_phon import sone_to_phon

from mosqito.sq_metrics.speech_intelligibility.sii_ansi.sii_ansi import sii_ansi
from mosqito.sq_metrics.speech_intelligibility.sii_ansi.sii_ansi_freq import sii_ansi_freq
from mosqito.sq_metrics.speech_intelligibility.sii_ansi.sii_ansi_level import sii_ansi_level



# -*- coding: utf-8 -*-

import numpy as np
from scipy.fft import fft, fftfreq

# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError(
        "In order to perform the tests you need the 'pytest' package.")
try:
    from SciDataTool import DataLinspace, DataTime
except ImportError:
    raise RuntimeError(
        "In order to handle Data objects you need the 'SciDataTool' package."
    )

# Local application imports
from mosqito.utils import load
from mosqito.sq_metrics import sharpness_din_st, sharpness_din_tv, sharpness_din_freq, sharpness_din_perseg


@pytest.fixture
def test_signal():
    # Input signal from DIN 45692_2009E
    sig, fs = load("tests/input/broadband_570.wav", wav_calib=1)
    sig_dict = {
        "signal": sig,
        "fs": fs,
        "S_din": 2.85,
    }
    return sig_dict


@pytest.mark.sharpness_din  # to skip or run sharpness test
def test_sharpness_din_st(test_signal):
    """Test function for the sharpness calculation of an audio signal
    The input signals come from DIN 45692_2009E. The compliance is assessed
    according to chapter 6 of the standard.
    Parameters
    ----------
    None
    Outputs
    -------
    None
    """

    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness = sharpness_din_st(sig, fs, weighting="din")

    # Check that the value is within the desired values +/- 5%
    # as per DIN 45692_2009E (chapter 6)
    np.testing.assert_allclose(sharpness, test_signal["S_din"], rtol=0.05)


@pytest.mark.sharpness_din  # to skip or run sharpness test
def test_sharpness_din_tv():
    """Test function for the sharpness calculation of an time-varying audio signal.

    Parameters
    ----------
    None
    Outputs
    -------
    None
    """

    # Input signal
    sig, fs = load(
        "tests/input/white_noise_442_1768_Hz_varying.wav", wav_calib=0.01)

    # Compute sharpness
    sharpness, time = sharpness_din_tv(sig, fs, weighting="din", skip=0.2)


@pytest.mark.sharpness_din
def test_sharpness_din_aures(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness = sharpness_din_st(sig, fs, weighting="aures")


@pytest.mark.sharpness_din
def test_sharpness_din_bismarck(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness = sharpness_din_st(sig, fs, weighting="bismarck")


@pytest.mark.sharpness_din
def test_sharpness_din_fastl(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness = sharpness_din_st(sig, fs, weighting="fastl")


@pytest.mark.sharpness_din  # to skip or run sharpness test
def test_sharpness_din_perseg(test_signal):
    """Test function for the sharpness calculation of an audio signal
    Test function for the function "comp_roughness" with 'din' method.
    The input signals come from DIN 45692_2009E. The compliance is assessed
    according to chapter 6 of the standard.
    Parameters
    ----------
    None
    Outputs
    -------
    None
    """

    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="aures")
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="bismarck")
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="fastl")
    sharpness, time_axis = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="din"
    )

    # Check that the value is within the desired values +/- 5%
    # as per DIN 45692_2009E (chapter 6)
    np.testing.assert_allclose(sharpness, test_signal["S_din"], rtol=0.05)


@pytest.mark.sharpness_din  # to skip or run sharpness test
def test_sharpness_din_perseg_sdt(test_signal):

    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]
    time = DataLinspace(
        name="time",
        unit="s",
        initial=0,
        final=(len(sig) - 1) / fs,
        number=len(sig),
        include_endpoint=True,
    )
    sig_data = DataTime(
        name="Test signal 5 (pinknoise 60 dB)",
        symbol="p",
        unit="Pa",
        axes=[time],
        values=sig,
    )
    # Compute sharpness
    sharpness, time_axis = sharpness_din_perseg(
        sig_data, fs, nperseg=2 ** 14, weighting="din", is_sdt_output=True
    )
    sharpness = sharpness.get_along('time')[sharpness.symbol]

    # Check that the value is within the desired values +/- 5%
    # as per DIN 45692_2009E (chapter 6)
    np.testing.assert_allclose(sharpness, test_signal["S_din"], rtol=0.05)


@pytest.mark.sharpness_din
def test_sharpness_din_perseg_aures(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="aures")


@ pytest.mark.sharpness_din
def test_sharpness_din_perseg_bismarck(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="bismarck")


@ pytest.mark.sharpness_din
def test_sharpness_din_perseg_fastl(test_signal):
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute sharpness
    sharpness, _ = sharpness_din_perseg(
        sig, fs, nperseg=2 ** 14, weighting="fastl")


@ pytest.mark.sharpness_din  # to skip or run sharpness test
def test_sharpness_din_freq(test_signal):
    """Test function for the sharpness calculation of an audio signal

    Test function for the function "comp_roughness" with 'din' method.
    The input signals come from DIN 45692_2009E. The compliance is assessed
    according to chapter 6 of the standard.
    One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """

    # Input signal from DIN 45692_2009E
    signal = {"data_file": "tests/input/broadband_570.wav",
              "S": test_signal["S_din"]}

    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]
    # Compute corresponding spectrum
    n = len(sig)
    spec = 2 / np.sqrt(2) / n * fft(sig)[0:n//2]
    freqs = fftfreq(n, 1/fs)[0:n//2]

    # Compute sharpness
    sharpness = sharpness_din_freq(spec, freqs, weighting="din")

    assert check_compliance(sharpness, signal)


def check_compliance(S, signal):
    """Check the comppiance of loudness calc. to ISO 532-1

    The compliance is assessed according to chapter 6 of the
    standard DIN 45692_2009E.

    Parameters
    ----------
    S : float
        computed sharpness value
    signal : dict
        {"data file" : <path to the standard file >
         "S" : <sharpness reference value>
         }

    Outputs
    -------
    tst : bool
        Compliance to the reference data
    """

    # Load reference value
    ref = signal["S"]

    # Test for DIN 45692_2009E comformance (chapter 6)
    tst = (S >= np.amax([ref * 0.95, ref - 0.05], axis=0)).all() and (
        S <= np.amin([ref * 1.05, ref + 0.05], axis=0)
    ).all()

    return tst


# test de la fonction
if __name__ == "__main__":
    # Reproduce the code from the fixture
    # Input signal from DIN 45692_2009E
    sig, fs = load("tests/input/broadband_570.wav", wav_calib=1)
    test_signal = {
        "signal": sig,
        "fs": fs,
        "S_din": 2.85,
    }
    # test_sharpness_din_st(test_signal)
    # test_sharpness_din_tv()
    # test_sharpness_din_freq(test_signal)
    # test_sharpness_din_perseg(test_signal)
    test_sharpness_din_perseg_sdt(test_signal)



# -*- coding: utf-8 -*-


from numpy import array, zeros, log10, maximum, where, sum

from mosqito.sq_metrics.speech_intelligibility.sii_ansi._band_procedure_data import _get_critical_band_data, _get_equal_critical_band_data, _get_octave_band_data, _get_third_octave_band_data
from mosqito.sq_metrics.speech_intelligibility.sii_ansi._speech_data import _get_critical_band_speech_data, _get_equal_critical_band_speech_data, _get_octave_band_speech_data, _get_third_octave_band_speech_data
from mosqito.utils.LTQ import LTQ
from mosqito.utils.conversion import freq2bark


def _main_sii(method, speech_spectrum, noise_spectrum, threshold):
    """Calculate core speech intelligibility index

    This function computes SII values based on ANSI S3.5 standard.

    Parameters
    ----------
    method: {"critical", "equally_critical", "third_octave", "octave"}
        Type of frequency band to be used for the calculation.
    speech_spectrum : array_like
        Speech spectrum [dB ref. 2e-5 Pa] with same size as the chosen method frequency axis.
    noise_spectrum : array_like
        Noise spectrum [dB ref. 2e-5 Pa] with same size as the chosen method frequency axis.
    threshold : array_like or 'zwicker'
        Threshold of hearing [dB ref. 2e-5 Pa] with same size as the chosen method frequency axis, or 'zwicker' to use the standard threshold.
        Default to None sets the threshold to zeros on each frequency band.
        
    Returns
    -------
    sii: numpy.ndarray
        Overall SII value.
    specific_sii: numpy.ndarray
        Specific SII values along the frequency axis.
    freq_axis: numpy.ndarray
        Frequency axis corresponding to the chosen method.
    """
    
    if (method!='critical') & (method!='equally_critical') & (method!='third_octave') & (method!='octave'):
        raise ValueError('Method should be within {"critical", "equally_critical", "third_octave", "octave"}.')
    
    # Get band data according to the chosen method
    if method == 'critical':
        CENTER_FREQUENCIES, LOWER_FREQUENCIES, UPPER_FREQUENCIES, IMPORTANCE, REFERENCE_INTERNAL_NOISE_SPECTRUM, STANDARD_SPEECH_SPECTRUM_NORMAL = _get_critical_band_data()
    elif method == 'equally_critical':
        CENTER_FREQUENCIES, LOWER_FREQUENCIES, UPPER_FREQUENCIES, IMPORTANCE, REFERENCE_INTERNAL_NOISE_SPECTRUM, STANDARD_SPEECH_SPECTRUM_NORMAL = _get_equal_critical_band_data()
    elif method == 'third_octave':
        CENTER_FREQUENCIES, _, _, _, IMPORTANCE, REFERENCE_INTERNAL_NOISE_SPECTRUM, STANDARD_SPEECH_SPECTRUM_NORMAL = _get_third_octave_band_data()
    elif method == 'octave':
        CENTER_FREQUENCIES, _, _, _, IMPORTANCE, REFERENCE_INTERNAL_NOISE_SPECTRUM, STANDARD_SPEECH_SPECTRUM_NORMAL = _get_octave_band_data()
    nbands = len(CENTER_FREQUENCIES)
    
    if threshold is None:
        T = zeros((nbands))
    elif threshold == 'zwicker':
        T = LTQ(freq2bark(CENTER_FREQUENCIES))
    else:
        T = array(threshold)
        
    # dB bandwidth adjustement
    if (method == 'critical_bands') or (method == 'equal_critical_bands'):
        noise_spectrum -= 10 * log10(UPPER_FREQUENCIES - LOWER_FREQUENCIES)

    # STEP 3
    if method == 'octave':
        Z = noise_spectrum
    else:
        V = speech_spectrum - 24
        B = maximum(noise_spectrum, V)
        if method == 'third_octave':
            C = -80 + 0.6 * (B + 10*log10(CENTER_FREQUENCIES)-6.353)
            Z = zeros((nbands))
            for i in range(nbands): 
                s = 0
                for k in range(i):
                    s += 10**(0.1*(B[k] + 3.32 * C[k] * log10(0.89 * CENTER_FREQUENCIES[i] / CENTER_FREQUENCIES[k])))
                Z[i] = 10 * log10(10**(0.1*noise_spectrum[i]) + s)
        else:
            C = -80 + 0.6 * (B + 10*log10(UPPER_FREQUENCIES - LOWER_FREQUENCIES))
            Z = zeros((nbands))
            for i in range(nbands): 
                s = 0
                for k in range(i-1):
                    s += 10**(0.1*(B[k] + 3.32 * C[k] * log10(CENTER_FREQUENCIES[i] / CENTER_FREQUENCIES[k])))                    
                Z[i] = 10 * log10(10**(0.1*noise_spectrum[i]) + s)
        # 4.3.2.4
        Z[0] = B[0]
    
    # STEP 4
    X = REFERENCE_INTERNAL_NOISE_SPECTRUM + T
    
    # STEP 5
    D = maximum(Z, X)
    
    # STEP 6
    L = 1 - (speech_spectrum - STANDARD_SPEECH_SPECTRUM_NORMAL -10)/160
    L[where(L>1)] = 1
    
    # STEP 7
    K = (speech_spectrum - D + 15)/30
    K[where(K>1)] = 1
    K[where(K<0)] = 0
    A = L * K
    
    # STEP 8
    SII = sum(IMPORTANCE * A)
    SII_specific = IMPORTANCE * A
    
    return SII, SII_specific, CENTER_FREQUENCIES





try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError(
        "In order to perform this validation you need the 'matplotlib' package."
        )


from scipy.special import comb
from scipy.signal import freqz
from numpy import (
    abs as np_abs,
    arange,
    exp,
    pi,
    log10,
    power as np_power,
    sqrt,
    insert as np_insert,
    sum as np_sum,
    array as np_array,
)


def _gammatone(freq, k=5, fs=48000, is_plot=False):
    """ECMA-418-2 Gammatone filter design

    This function computes the coefficients of a gammatone digital filter according to ECMA-418-2 section 5.1.3.

    Parameters
    ----------
    freq: float
        Center frequency of the filter ['Hz'].

    k: int, optional
        The order of the filter. Default is "5" according to ECMA-418.2.

    fs: float, optional
        The sampling frequency of the signal. Default is 48000 Hz.

    Returns
    -------
    bm_prim, am_prim: ndarray, ndarray
        Numerator (b) and denominator (a) polynomials of the filter.

    """

    # ECMA-418-2 constants
    af_f0 = 81.9289
    c = 0.1618

    # Bandwidth (ECMA 418-2 equation 7)
    delta_f = sqrt((af_f0 ** 2) + ((c * freq) ** 2))

    # Time constant, delay (ECMA 418-2 equation 5)
    binom = comb(2 * k - 2, k - 1, exact=True)
    tau = (1 / (2 ** (2 * k - 1))) * binom * (1.0 / delta_f)

    # "d" coefficient
    d = exp(-1 / (fs * tau))

    # coeff am (ECMA 418-2 equation 11)
    m = arange(5) + 1
    am = np_power((-d), m) * comb(5, m)
    am = np_insert(am, 0, 1)

    # coeff bm (ECMA 418-2 equation 12)
    em = np_array([0, 1, 11, 11, 1])
    i = arange(4) + 1
    denom = np_sum(em[i] * d ** i)
    m = arange(5)
    bm = ((1 - d) ** k) / denom * (d ** m) * em[m]

    # band pass filter coefficients (ECMA 418-2 equation 13 & 14)
    # [by modifying the filter ceofficients with a negative exponential,
    # the filter is a low-pass filter instead of the expected bandpass
    # filter]
    m = arange(6)
    exponential = exp(-1j * 2 * pi * freq * m / fs)
    am_prim_ecma = am * exponential
    bm_prim_ecma = bm * exponential[:-1]

    # band pass filter coefficients (ECMA 418-2 from equation 13 & 14)
    # [corrected to get a bandpass filter, to be validated]
    m = arange(6)
    exponential = exp(1j * 2 * pi * freq * m / fs)
    am_prim = am * exponential
    bm_prim = bm * exponential[:-1]

    if is_plot:
        w, h = freqz(bm, am, worN=round(fs / 2), fs=fs)
        h_db = 20.0 * log10(np_abs(h))
        plt.semilogx(w, h_db, label="am, bm from eq. 11 and 12")

        w, h = freqz(bm_prim_ecma, am_prim_ecma, worN=round(fs / 2), fs=fs)
        h_db = 20.0 * log10(np_abs(h))
        plt.semilogx(w, h_db, label="am', bm' from eq. 13 and 14")

        w, h = freqz(bm_prim, am_prim, worN=round(fs / 2), fs=fs)
        h_db = 20.0 * log10(np_abs(h))
        plt.semilogx(w, h_db, label="am', bm' with positive exp")

        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Amplitude [dB]")
        plt.grid(which="both", axis="both")
        plt.legend()

    return bm_prim, am_prim


import numpy as np

from mosqito.sound_level_meter.noct_spectrum._nominal_frequency import (
    NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
)


def _center_freq(fmin, fmax, n=3, G=10, fr=1000):
    """
    Compute nth octave filter frequencies

    References:
        ANSI S1.1-1986 (ASA 65-1986): Specifications for
        Octave-Band and Fractional-Octave-Band Analog and
        Digital Filters.

    Parameters
    ----------
    fmin : float
        Min frequency band [Hz]
    fmax : float
        Max frequency band [Hz]
    n : int, optional
        number of bands pr octave. Default to 3
    G : int, optional
        System for specifying the exact geometric mean frequencies.
        Can be base 2 or base 10. Default to base 10
    fr : int, optional
        Reference frequency. Shall be set to 1 kHz for audible frequency
        range, to 1 Hz for infrasonic range (f < 20 Hz) and to 1 MHz for
        ultrasonic range (f > 31.5 kHz). Default to 1000 Hz

    Outputs
    -------
    f_exact : ndarray
        Exact center frequencies
    """

    # determine band numbers
    b = 1 / n
    if G == 2:
        U = 2 ** b  # ANSI eq2
    elif G == 10:
        U = 10 ** (3 * b / 10)  # ANSI eq4
    else:
        raise ValueError(
            """ERROR: Only base 2 and base 10 are allowed for nth
            octave center frequency definition"""
        )
    [kmin, kmax] = np.round(
        np.log10(np.array([fmin, fmax]) / fr) / np.log10(U), 0)

    # Band numbers such that f_exact = fr for k=0
    k = np.arange(kmin, kmax + 1).astype(int)

    # compute ANSI eq1
    f_exact = fr * U ** k

    ####
    # get normalized frequencies
    ####

    f_nom = f_exact.copy()

    if n == 1:
        freq = NOMINAL_OCTAVE_CENTER_FREQUENCIES
    elif n == 3:
        freq = NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES
    if n == 1 or n == 3:
        i_ref = np.where(freq == fr)[0][0]
        ind = np.where((k >= -i_ref) & (k < (len(freq) - i_ref)))
        f_nom = freq[k[ind] + i_ref]

    # TODO
    # Manage other values of n
    # Manage band outside the known bands

    return f_exact, f_nom


if __name__ == "__main__":
    f_exact, f_nom = _center_freq(25, 50000, n=3, G=10, fr=1000)
    pass



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 11 09:04:04 2020

@author: pc
"""
import numpy as np


def _critical_band(f0):
    """Analytical definition of the critical band centered on f0
    according to ECMA-74 annex D.8"""

    delta_fc = 25 + 75 * np.power(1 + 1.4 * np.power(f0 / 1000, 2), 0.69)

    if f0 < 500:
        f1 = f0 - delta_fc / 2
        f2 = f0 + delta_fc / 2

    elif f0 >= 500:
        f1 = -1 * delta_fc / 2 + np.sqrt(delta_fc ** 2 + 4 * f0 ** 2) / 2
        f2 = f1 + delta_fc

    return f1, f2


def _lower_critical_band(f0):
    """Analytical definition of the critical band immediately below and contiguous
    with the critical band centered on f0 according to ECMA-74 annex D.10
    """

    f2, _ = _critical_band(f0)

    if f0 < 171.4:
        c0 = 20
        c1 = 0
        c2 = 0
    elif f0 >= 171.4 and f0 <= 1600:
        c0 = -149.5
        c1 = 1.001
        c2 = -6.9e-05
    elif f0 > 1600:
        c0 = 6.8
        c1 = 0.806
        c2 = -8.2e-06

    f1 = c0 + c1 * f0 + c2 * f0 ** 2

    return f1, f2


def _upper_critical_band(f0):
    """Analytical definition of the critical band immediately above and contiguous
    with the critical band centered on f0 according to ECMA-74 annex D.10
    """

    _, f1 = _critical_band(f0)

    if f0 <= 1600:
        c0 = 149.5
        c1 = 1.035
        c2 = 7.7e-05
    elif f0 > 1600:
        c0 = 3.3
        c1 = 1.215
        c2 = 2.16e-05

    f2 = c0 + c1 * f0 + c2 * f0 ** 2

    return f1, f2



# -*- coding: utf-8 -*-

# Standard imports
import numpy as np
from numpy.fft import fft, ifft
import math

# Local imports
from mosqito.utils.LTQ import LTQ
from mosqito.sq_metrics.roughness.roughness_dw._ear_filter_coeff import (
    _ear_filter_coeff,
)
from mosqito.utils.conversion import freq2bark, db2amp, amp2db, bark2freq

def _roughness_dw_main_calc(spec, freq_axis, fs, gzi, hWeight):
    """
    Daniel and Weber roughness main calculation

    Parameters
    ----------
    spec : array
        An amplitude or complex spectrum.
    freq_axis : array
        Frequency axis in [Hz].
    fs : integer
        Sampling frequency.
    gzi : array
        gzi weighting function.
    hWeight : array
        H weighting function.

    Returns
    -------
    R : float
        Roughness computed for the given spectrum.

    """
    if len(spec) != len(freq_axis):
        raise ValueError(
            "spectrum and frequency axis should have the same number of points !"
        )

    # convert spectrum to 2-sided
    spec = np.concatenate((spec, spec[len(spec)::-1]))

    n = len(spec)
    # Frequency axis in Bark
    bark_axis = freq2bark(freq_axis)
    # Highest frequency
    nZ = np.arange(1, n//2 + 1, 1)

    # Calculate Zwicker a0 factor (transfer characteristic of the outer and inner ear)
    a0 = np.zeros((n))
    a0[nZ - 1] = db2amp(_ear_filter_coeff(bark_axis), ref=1)
    spec = a0 * spec

    # Conversion of the spec into dB
    module = np.abs(spec[0:n//2])
    spec_dB = amp2db(module, ref=2e-5)
    
    # Find the audible components within the spec
    threshold = LTQ(bark_axis, reference="roughness")
    audible_index = np.where(spec_dB > threshold)[0]
    # Number of audible frequencies
    n_aud = len(audible_index)

    # --------------------------------- stage 1 ------------------------------------
    # ----------------Creation of the specific excitations functions----------------

    # Terhardt's slopes definition
    # lower slope [dB/Bark]
    s1 = -27
    s2 = np.zeros((n_aud))
    # upper slope [dB/Bark]
    for k in np.arange(0, n_aud, 1):
        s2[k] = min(
            -24 - (230 / freq_axis[audible_index[k]]) + (0.2 * spec_dB[audible_index[k]]),
            0,
        )

    # The excitation pattern are calculated for 47 overlapping 1-bark-wide channels
    n_channel = 47
    # Channels number
    zi = np.arange(1, n_channel + 1) / 2
    # Center frequencies for each channel
    zb = bark2freq(zi) * n / fs
    # Minimum excitation level
    minExcitDB = np.interp(zb, nZ, threshold)
    
    ch_low = np.zeros((n_aud))
    ch_high = np.zeros((n_aud))
    for i in np.arange(0, n_aud):
        # Lower limit of the channel corresponding to each component
        ch_low[i] = math.floor(2 * bark_axis[audible_index[i]]) - 1
        # Higher limit
        ch_high[i] = math.ceil(2 * bark_axis[audible_index[i]]) - 1

    # Creation of the excitation pattern
    slopes = np.zeros((n_aud, n_channel))
    for k in np.arange(0, n_aud):
        levDB = spec_dB[audible_index[k]]
        b = bark_axis[audible_index[k]]
        for j in np.arange(0, int(ch_low[k] + 1)):
            sl = (s1 * (b - ((j + 1) * 0.5))) + levDB
            if sl > minExcitDB[j]:
                slopes[k, j] = db2amp(sl, ref=0.00002)
        for j in np.arange(int(ch_high[k]), n_channel):
            sl = (s2[k] * (((j + 1) * 0.5) - b)) + levDB
            if sl > minExcitDB[j]:
                slopes[k, j] = db2amp(sl, ref=0.00002)

    # Initialization
    hBP = np.zeros((n_channel, n))
    mod_depth = np.zeros((n_channel))

    # Definition of the excitation amplitude
    for i in np.arange(0, n_channel, 1):
        exc = np.zeros((n), dtype=complex)
        for j in np.arange(0, n_aud, 1):
            ind = audible_index[j]
            # the component belongs to the bark window
            if ch_low[j] == i:
                ampl = 1
            elif ch_high[j] == i:
                ampl = 1
            # the component is higher than the bark window
            elif ch_high[j] > i:
                ampl = slopes[j, i + 1] / module[ind]
            # the component is lower than the considered window
            else:
                ampl = slopes[j, i - 1] / module[ind]

            # reconstruction of the spec
            exc[ind] = ampl * spec[ind]

        # The temporal specific excitation functions are obtained by IFFT
        temporal_excitation = np.abs(n * np.real(ifft(exc)))
        # ------------------------------- stage 2 --------------------------------------
        # ---------------------modulation depth calculation-----------------------------

        # The fluctuations of the envelope are contained in the low frequency part
        # of the spec of specific excitations in absolute value
        h0 = np.mean(temporal_excitation)
        envelope_spec = fft(temporal_excitation - h0)

        # This spec is weighted to model the low-frequency  bandpass
        # characteristic of the roughness on modulation frequency
        envelope_spec = envelope_spec * hWeight[i, :]
        # The time functions of the bandpass filtered envelopes hBPi(t)
        # are calculated via inverse Fourier transform :
        hBP[i, :] = 2 * np.real(ifft(envelope_spec))

        # Modulation depth estimation is given by envelope RMS values
        # and excitation functions time average :
        hBPrms = np.sqrt(np.mean(np.power(hBP[i, :], 2)))
        if h0 > 0:
            mod_depth[i] = hBPrms / h0
            if mod_depth[i] > 1:
                mod_depth[i] = 1
        else:
            mod_depth[i] = 0
    # ------------------------------- stage 3 --------------------------------------
    # ----------------roughness calculation with cross correlation------------------

    # Crosscorrelation coefficients between the envelopes of the channels
    # i and i+2 with dz= 1 bark
    ki = np.zeros((47))

    for i in range(0, 45):
        if hBP[i].all() != 0 and hBP[i + 2].all() != 0:
            ki[i] = np.corrcoef(hBP[i, :], hBP[i + 2, :])[0, 1]

    # Specific roughness calculation with gzi the modulation depth weighting
    # function given by Aures
    R_spec = np.zeros((47))

    R_spec[0] = gzi[0] * pow(mod_depth[0] * ki[0], 2)
    R_spec[1] = gzi[1] * pow(mod_depth[1] * ki[1], 2)
    for i in np.arange(2, 45):
        R_spec[i] = gzi[i] * pow(mod_depth[i] * ki[i] * ki[i - 2], 2)
    R_spec[45] = gzi[45] * pow(mod_depth[45] * ki[43], 2)
    R_spec[46] = gzi[46] * pow(mod_depth[46] * ki[44], 2)

    # Total roughness calculation with calibration factor of 0.25 given in the article
    # to produce a roughness of 1 asper for a 1-kHz, 60dB tone with carrier frequency
    # of 70 Hz and a modulation depth of 1

    R = 0.25 * sum(R_spec)

    return R, R_spec, zi






# -*- coding: utf-8 -*-
"""
@author: Daniel Jiménez-Caminero Costa
"""

import numpy as np


def equal_loudness_contours(phones):
    """ "This function will return a 29-point equal loudness contour for your desired phon level. The frequencies
    evaluated in this function only span from 20Hz - 12.5kHz, and only 29 selective frequencies are covered. This is
    the limitation of the ISO standard".

    Function based on the MATLAB project from Jeff Tackett, 'ISO 226 Equal-Loudness-Level Contour Signal'
    (es.mathworks.com/matlabcentral/fileexchange/7028-iso-226-equal-loudness-level-contour-signal).

    Parameters
    ----------
    phones: int
        Logarithmic unit for the evaluation curve of the perceived sound (1phon = 1dB @ 1kHz).

    Returns
    -------
    spl_array: numpy.array
        Sound pressure level ('SPL') values.

    frequencies_array: numpy.array
        Frequencies in which the function computes the resulting 'SPL' values.
    """
    n_frequencies = 29
    freq_array = np.array([20.0, 25.0, 31.5, 40.0, 50.0, 63.0, 80.0, 100.0, 125.0, 160.0, 200.0, 250.0, 315.0, 400.0,
                           500.0, 630.0, 800.0, 1000.0, 1250.0, 1600.0, 2000.0, 2500.0, 3150.0, 4000.0, 5000.0, 6300.0,
                           8000.0, 10000.0, 12500.0])

    af = [0.532, 0.506, 0.480, 0.455, 0.432, 0.409, 0.387, 0.367, 0.349, 0.330, 0.315, 0.301, 0.288, 0.276, 0.267,
          0.259, 0.253, 0.250, 0.246, 0.244, 0.243, 0.243, 0.243, 0.242, 0.242, 0.245, 0.254, 0.271, 0.301]

    Lu = [-31.6, -27.2, -23.0, -19.1, -15.9, -13.0, -10.3, -8.1, -6.2, -4.5, -3.1, -2.0, -1.1, -0.4, 0.0, 0.3, 0.5, 0.0,
          -2.7, -4.1, -1.0, 1.7, 2.5, 1.2, -2.1, -7.1, -11.2, -10.7, -3.1]

    Tf = [78.5, 68.7, 59.5, 51.1, 44.0, 37.5, 31.5, 26.5, 22.1, 17.9, 14.4, 11.4, 8.6, 6.2, 4.4, 3.0, 2.2, 2.4, 3.5,
          1.7, -1.3, -4.2, -6.0, -5.4, -1.5, 6.0, 12.6, 13.9, 12.3]

    # Ln = phones
    # Lp = SPL
    spl_array = np.zeros(n_frequencies, dtype=float)

    # Deriving sound pressure level from loudness level (iso226 sect 4.1). Ln = phones
    # An upgrade of this section could be to do all the calculations in one run by using numpy functions
    for i in range(n_frequencies):
        # "Af" is the argument of the conversion formula
        Af = (4.47 * (10 ** (-3))) * ((10 ** (0.025 * phones)) - 1.15) + \
             ((0.4 * (10 ** (((Tf[i] + Lu[i]) / 10) - 9))) ** af[i])

        spl_array[i] = ((10. / af[i]) * np.log10(Af)) - Lu[i] + 94

    return spl_array, freq_array



# -*- coding: utf-8 -*-

import numpy as np
from numpy.fft import fft

# local function import
from mosqito.utils.conversion import amp2db


def spectrum(signal,fs, nfft='default', window='hanning', one_sided=True, db=True):
    """
    Compute one-sided spectrum from a time signal in Pa.

    Parameters
    ----------
    signal : np.array
        A time signal [nperseg x nseg].
    fs : integer
        Sampling frequency.
    db : boolean, optional
        Indicates if the spectrum is in dB values. Default is True.

    Returns
    -------
    spectrum : np.array
        Spectrum [freq_axis x nseg].
    freq_axis : np.array
        Frequency axis.

    """
    if len(signal.shape)>1:
        nseg = signal.shape[1]
    else:
        nseg = 1
    
    # Number of points for the fft
    if nfft == 'default':
        if len(signal.shape) == 1:
            nfft = len(signal)
        else :
            nfft = signal.shape[0]

    # Window definition
    if window == 'hanning':
        window = np.hanning(nfft)
    elif window == 'blackman':
        window = np.blackman(nfft)
        
    # Amplitude correction
    window = window / np.sum(window)
    
    if len(signal.shape)>1:
        window = np.tile(window,(nseg,1)).T
    
    # Creation of the spectrum by FFT
    if one_sided == True:
        spectrum = fft(signal * window, n=nfft, axis=0)[0:nfft//2] * 1.42
        freq_axis = np.arange(1, nfft//2+1, 1) * (fs / nfft)
    else:
        spectrum = fft(signal * window, n=nfft, axis=0) * 1.42
        freq_axis = np.concatenate((np.arange(1, nfft//2+1, 1) * (fs / nfft), np.arange(nfft//2+1, 1, -1) * (fs / nfft)))

    if db == True:
        # Conversion into dB level
        module = np.abs(spectrum)
        spectrum = amp2db(module, ref=0.00002)
    
    return spectrum, freq_axis


# -*- coding: utf-8 -*-

# Standard library imports
import numpy as np

# Local applications imports
from mosqito.sq_metrics.loudness.loudness_zwst._main_loudness import _main_loudness
from mosqito.sq_metrics.loudness.loudness_zwst._calc_slopes import _calc_slopes
from mosqito.sq_metrics.loudness.loudness_zwtv._nonlinear_decay import _nl_loudness
from mosqito.sq_metrics.loudness.loudness_zwtv._temporal_weighting import _temporal_weighting
from mosqito.sq_metrics.loudness.loudness_zwtv._third_octave_levels import _third_octave_levels


def loudness_zwtv(signal, fs, field_type='free'):
    """Calculate Zwicker-loudness for time-varying signals
    Calculate the acoustic loudness according to Zwicker method for
    time-varying signals.
    Normatice reference:
        DIN 45631/A1:2010
        ISO 532-1:2017 (method 2)
    The code is based on C program source code published alongside
    with ISO 532-1 standard.
    Note that for reasons of normative continuity, as defined in the
    preceeding standards, the method is in accordance with
    ISO 226:1987 equal loudness contours (instead of ISO 226:2003)
    Parameters
    ----------
    signal : numpy.array
        A time signal values [Pa].
    fs : integer
        Sampling frequency.
    field_type : str
        Type of soundfield corresponding to signal ("free" by or "diffuse").
        
    Outputs
    -------
    N : float
        Calculated loudness [sones], size (Ntime,).
    N_specific : numpy.ndarray
        Specific loudness [sones/bark], size (Nbark, Ntime).
    bark_axis : numpy.ndarray
        Corresponding bark axis, size (Nbark,).
    time_axis : numpy.ndarray
        Time axis, size (Ntime,).

    """        

    # Compute third octave band spectrum vs. time
    spec_third, time_axis, _ = _third_octave_levels(signal, fs)

    # Calculate core loudness (vectorized version)
    core_loudness = _main_loudness(spec_third, field_type)

    #
    # Nonlinearity
    core_loudness = _nl_loudness(core_loudness)
    #
    # Calculation of specific loudness
    loudness, spec_loudness = _calc_slopes(core_loudness)

    # temporal weigthing
    filt_loudness = _temporal_weighting(loudness)
    #
    # Decimation from temporal resolution 0.5 ms to 2ms and return
    dec_factor = 4
    N = filt_loudness[::dec_factor]
    N_spec = spec_loudness[:, ::dec_factor]
    time_axis = time_axis[::dec_factor]
    #
    # Build bark axis
    bark_axis = np.linspace(0.1, 24, int(24 / 0.1))

    return N, N_spec, bark_axis, time_axis



# -*- coding: utf-8 -*-
"""
@author: Daniel Jiménez-Caminero Costa
"""

import numpy as np


def phone2spl(phones, tf, af, lu):
    """ Conversion from phones to Sound Pressure Level (dB SPL), based on the standard ISO/FDIS 226:2003.

    Parameters
    ----------
    phones: float
        Logarithmic unit of loudness ('phon' = Ln).

    tf: float
        Threshold of hearing.

    af: float
        Exponent used for loudness perception.

    lu: float
        Factor of the linear transfer function normalized at 1 kHz.

    Returns
    -------
    spl: float
        Sound pressure level ('dB SPL').

    """
    # "Af" is argument of the dB SPL conversion formula
    Af = (4.47 * (10 ** (-3))) * ((10 ** (0.025 * phones)) - 1.15) + ((0.4 * (10 ** (((tf + lu) / 10) - 9))) ** af)

    # Result of the phone value converted to dB SPL value
    spl = ((10. / af) * np.log10(Af)) - lu + 94

    return spl



# -*- coding: utf-8 -*-

# Optional package import
try:
    import pytest
except ImportError:
    raise RuntimeError("In order to perform the tests you need the 'pytest' package.")
try:
    from SciDataTool import DataLinspace, DataTime
except ImportError:
    raise RuntimeError(
        "In order to handle Data objects you need the 'SciDataTool' package."
    )


import numpy as np
from numpy.fft import fft, fftfreq

# Local application imports
from mosqito.utils import load, isoclose
from mosqito.sq_metrics import loudness_zwst, loudness_zwst_freq, loudness_zwst_perseg
from mosqito.sq_metrics.loudness.loudness_zwst._main_loudness import _main_loudness
from mosqito.sq_metrics.loudness.loudness_zwst._calc_slopes import _calc_slopes
from tests.input.Test_signal_1 import test_signal_1


@pytest.fixture
def test_signal():
    sig, fs = load(
        "tests/input/Test signal 5 (pinknoise 60 dB).wav", wav_calib=2 * 2 ** 0.5
    )
    N_specif_iso = np.genfromtxt("tests/input/test_signal_5.csv", skip_header=1)
    sig_dict = {
        "signal": sig,
        "fs": fs,
        "N_iso": 10.498,
        "N_specif_iso": N_specif_iso,
    }
    return sig_dict


@pytest.mark.loudness_zwst  # to skip or run only loudness zwicker stationary tests
def test_loudness_zwst_3oct():
    """Test function for the script loudness_zwst

    Test function for the script loudness_zwicker_stationary with
    third octave band spectrum as input. The input spectrum is
    provided by ISO 532-1 annex B2, the compliance is assessed
    according to section 5.1 of the standard. One .png compliance
    plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """

    # Target values
    N_iso = 83.296
    N_specif_iso = np.genfromtxt("tests/input/test_signal_1.csv", skip_header=1)

    # Compute loudness
    Nm = _main_loudness(test_signal_1, field_type="free")
    N, N_specific = _calc_slopes(Nm)

    # Assert compliance
    is_isoclose_N = isoclose(N_iso, N, rtol=5 / 100, atol=0.1)
    is_isoclose_N_specific = isoclose(N_specific, N_specif_iso, rtol=5 / 100, atol=0.1)
    assert is_isoclose_N and is_isoclose_N_specific


@pytest.mark.loudness_zwst
def test_loudness_zwst_wav(test_signal):
    """Test function for the script loudness_zwicker_stationary

    Test function for the script loudness_zwicker_stationary with
    .wav file as input. The input file is provided by ISO 532-1 annex
    B3, the compliance is assessed according to section 5.1 of the
    standard. One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    # Test signal as input for stationary loudness
    # (from ISO 532-1 annex B3)
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute Loudness
    N, N_specific, bark_axis = loudness_zwst(sig, fs)

    # Assert compliance
    is_isoclose_N = isoclose(N, test_signal["N_iso"], rtol=5 / 100, atol=0.1)
    is_isoclose_N_specific = isoclose(
        N_specific, test_signal["N_specif_iso"], rtol=5 / 100, atol=0.1
    )
    assert is_isoclose_N and is_isoclose_N_specific


@pytest.mark.loudness_zwst
def test_loudness_zwst_44100Hz():
    """Test function for the script loudness_zwicker_stationary
    with input .wav file sampled at 44.1 kHz
    """
    # Test signal as input for stationary loudness
    # (from ISO 532-1 annex B3) resampled
    sig, fs = load(
        "tests/input/Test signal 3 (1 kHz 60 dB)_44100Hz.wav", wav_calib=2 * 2 ** 0.5
    )

    # Target values
    N_iso = 4.019
    N_specif_iso = np.genfromtxt("tests/input/test_signal_3.csv", skip_header=1)

    # Compute Loudness
    N, N_specific, bark_axis = loudness_zwst(sig, fs)

    # Assert compliance
    is_isoclose_N = isoclose(N, N_iso, rtol=5 / 100, atol=0.1)
    is_isoclose_N_specific = isoclose(N_specific, N_specif_iso, rtol=5 / 100, atol=0.1)
    assert is_isoclose_N and is_isoclose_N_specific


@pytest.mark.loudness_zwst
def test_loudness_zwst_perseg(test_signal):
    sig = test_signal["signal"]
    fs = test_signal["fs"]

    # Compute Loudness
    N, N_specific, bark_axis, time_axis = loudness_zwst_perseg(
        sig, fs, nperseg=8192 * 2, noverlap=4096
    )

    # Check that all values are within the desired values +/- 5%
    np.testing.assert_allclose(N, 10.498, rtol=0.05)


@pytest.mark.loudness_zwst
def test_loudness_zwst_sdt(test_signal):
    sig = test_signal["signal"]
    fs = test_signal["fs"]
    time = DataLinspace(
        name="time",
        unit="s",
        initial=0,
        final=(len(sig) - 1) / fs,
        number=len(sig),
        include_endpoint=True,
    )
    sig_data = DataTime(
        name="Test signal 5 (pinknoise 60 dB)",
        symbol="p",
        unit="Pa",
        axes=[time],
        values=sig,
    )
    N, N_specific, _ = loudness_zwst(sig_data, fs, is_sdt_output=True)
    N_specific = N_specific.get_along("Critical band rate")[N_specific.symbol]

    # Assert compliance
    is_isoclose_N = isoclose(N, test_signal["N_iso"], rtol=5 / 100, atol=0.1)
    is_isoclose_N_specific = isoclose(
        N_specific, test_signal["N_specif_iso"], rtol=5 / 100, atol=0.1
    )
    assert is_isoclose_N and is_isoclose_N_specific


@pytest.mark.loudness_zwst
def test_loudness_zwst_perseg_sdt(test_signal):
    sig = test_signal["signal"]
    fs = test_signal["fs"]
    time = DataLinspace(
        name="time",
        unit="s",
        initial=0,
        final=(len(sig) - 1) / fs,
        number=len(sig),
        include_endpoint=True,
    )
    sig_data = DataTime(
        name="Test signal 5 (pinknoise 60 dB)",
        symbol="p",
        unit="Pa",
        axes=[time],
        values=sig,
    )
    # Compute Loudness
    N, N_specific, bark_axis, time_axis = loudness_zwst_perseg(
        sig_data, fs, nperseg=8192 * 2, noverlap=4096, is_sdt_output=True
    )
    N = N.get_along("time")[N.symbol]

    # Check that all values are within the desired values +/- 5%
    np.testing.assert_allclose(N, 10.498, rtol=0.05)


@pytest.mark.loudness_zwst  # to skip or run only loudness zwicker stationary tests
def test_loudness_zwst_freq(test_signal):
    """Test function for the script loudness_zwicker_stationary

    Test function for the script loudness_zwicker_stationary with
    complex spectrum file as input. The input file is provided by ISO 532-1 annex
    B3, the compliance is assessed according to section 5.1 of the
    standard. One .png compliance plot is generated.

    Parameters
    ----------
    None

    Outputs
    -------
    None
    """
    # Input signal
    sig = test_signal["signal"]
    fs = test_signal["fs"]
    # Compute corresponding spectrum
    n = len(sig)
    spec = np.abs(2 / np.sqrt(2) / n * fft(sig)[0 : n // 2])
    freqs = fftfreq(n, 1 / fs)[0 : n // 2]
    # Compute Loudness
    N, N_specific, bark_axis = loudness_zwst_freq(spec, freqs)

    # 2D inputs
    spec = np.tile(spec, (4, 1)).T
    N1, N1_specific, bark_axis = loudness_zwst_freq(spec, freqs)

    freqs = np.tile(freqs, (4, 1)).T
    N2, N2_specific, bark_axis = loudness_zwst_freq(spec, freqs)

    # Assert compliance
    is_isoclose_N = isoclose(N, test_signal["N_iso"], rtol=5 / 100, atol=0.1)
    is_isoclose_N_specific = isoclose(
        N_specific, test_signal["N_specif_iso"], rtol=5 / 100, atol=0.1
    )
    assert is_isoclose_N and is_isoclose_N_specific


@pytest.mark.loudness_zwst
def test_loudness_zwst_gi0():
    """
    Test with a signal creating some zero values in gi variable
    of _main_loudness function (bug solved)
    """

    sig = np.load("tests/input/test_signal_gi0.npy")
    fs = 48000

    # Compute Loudness
    N, N_specific, bark_axis, time_axis = loudness_zwst_perseg(
        sig, fs, nperseg=8192 * 2, noverlap=4096
    )


# test
if __name__ == "__main__":
    # Reproduce the code from the fixture
    sig, fs = load(
        "tests/input/Test signal 5 (pinknoise 60 dB).wav", wav_calib=2 * 2 ** 0.5
    )
    N_specif_iso = np.genfromtxt("tests/input/test_signal_5.csv", skip_header=1)
    test_signal = {
        "signal": sig,
        "fs": fs,
        "N_iso": 10.498,
        "N_specif_iso": N_specif_iso,
    }

    # test_loudness_zwst_3oct()
    # test_loudness_zwst_wav(test_signal)
    # test_loudness_zwst_44100Hz()
    # test_loudness_zwst_perseg(test_signal)
    # test_loudness_zwst_sdt(test_signal)
    # test_loudness_zwst_perseg_sdt(test_signal)
    # test_loudness_zwst_freq(test_signal)
    test_loudness_zwst_gi0()



# -*- coding: utf-8 -*-

# Standard library imports
import numpy as np
from scipy.signal import decimate, butter, sosfilt


def _n_oct_time_filter(sig, fs, fc, alpha, N=3):
    """Design of a nth octave filter set

    Designs a digital 1/3-octave filter with center frequency fc for
    sampling frequency fs. The filter is designed according to the
    Order-N specification of the ANSI S1.1-1986 standard. Default
    value for N is 3.

    References:
        ANSI S1.1-1986 (ASA 65-1986): Specifications for
        Octave-Band and Fractional-Octave-Band Analog and
        Digital Filters.

    Parameters
    ----------
    sig : numpy.ndarray
        Time signal [any unit]
    fs : float
        Sampling frequency [Hz]
    fc : float
        Filter exact center frequency [Hz]
    alpha : float
        Ratio of the upper and lower band-edge frequencies to the mid-band
        frequency
    n : int, optional
        Filter order. Default to 3

    Outputs
    -------
    level : float
        Rms level of sig in the third octave band centered on fc
    """

    # Check for Nyquist-Shannon criteria
    if fc > 0.88 * (fs / 2):
        raise ValueError(
            """ERROR: Design not possible. Filter center frequency shall
            verify: fc <= 0.88 * (fs / 2)"""
        )

    # Check for high fc/fs causing filter design issue
    # [ref needed] and downsample if needed
    if fc < fs / 200:
        q = 2
        while fc < fs / q / 200:
            q += 1
        sig = decimate(sig, q, axis=0)
        fs = fs / q

    # Normalized cutoff frequencies
    w1 = fc / (fs / 2) / alpha
    w2 = fc / (fs / 2) * alpha

    # define filter coefficient
    sos = butter(int(N), (w1, w2), "bandpass", analog=False, output='sos')

    # filter signal
    sig_filt = sosfilt(sos, sig, axis=0)

    # Compute overall rms level
    level = np.sqrt(np.mean(sig_filt ** 2, axis=0))

    return level


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 21 16:44:36 2020

@author: wantysal
"""
# Standard library import
import numpy as np

# Local import
from mosqito.sound_level_meter.noct_spectrum._getFrequencies import _getFrequencies


def _spectrum_smoothing(freqs_in, spec, noct, low_freq, high_freq, freqs_out):
    """
        Compute smoothed spectrum according to the n-th octave band chosen

    Parameters
    ----------
    freqs : numpy.array
        frequency axis dim (nperseg)
    spec : numpy.array
        spectrum in dB (nperseg, nseg)
    noct : integer
        n-th octave-band according to which smooth the spectrum
    low_freq : float
        lowest frequency of the n-th octave bands
    high_freq : float
        highest frequency of the n-th octave bands
    freqs_out : numpy.array
        frequency axis along which the smoothed spectrum is given

    Returns
    -------
    smoothed-spectrum : numpy.array
        smoothed spectrum along the given frequency axis

    """
    nperseg = spec.shape[0]
    if len(spec.shape) > 1:
        nseg = spec.shape[1]
    else:
        nseg = 1
    spec = spec.ravel()
    stop = np.arange(1, nseg + 1) * nperseg
    freqs_in = freqs_in.ravel()

    # n-th octave bands filter
    filter_freqs = _getFrequencies(low_freq, high_freq, noct, G=10, fr=1000)["f"]
    filter_freqs[len(filter_freqs) - 1, 2] = high_freq
    filter_freqs[0, 0] = low_freq

    # Smoothed spectrum creation
    nb_bands = filter_freqs.shape[0]
    smoothed_spectrum = np.zeros((nb_bands, nseg))
    i = 0
    # Each band is considered individually until all of them have been treated
    while nb_bands > 0:
        # Find the index of the spectral components within the frequency bin
        bin_index = np.where((freqs_in >= filter_freqs[i, 0]) & (freqs_in <= filter_freqs[i, 2]))[0]

        # If the frequency bin is empty, it is deleted from the list
        if len(bin_index) == 0:
            smoothed_spectrum = np.delete(smoothed_spectrum, i, axis=1)
            filter_freqs = np.delete(filter_freqs, i, axis=0)
            nb_bands -= 1

        else:
            # The spectral components within the frequency bin are averaged on an energy basis
            spec_sum = np.zeros((nseg))
            for j in range(nseg):
                if len(bin_index[(bin_index < stop[j]) & (bin_index > (stop[j] - nperseg))])!= 0:
                    spec_sum[j] = np.mean(10** (spec[bin_index[(bin_index < stop[j]) & (bin_index > (stop[j] - nperseg))]]/ 10))
                else:
                    spec_sum[j] = 1e-12
            smoothed_spectrum[i, :] = 10 * np.log10(spec_sum)# / len(bin_index[(bin_index < stop[j]) & (bin_index > (stop[j] - m))]))
        nb_bands -= 1
        i += 1
    # Pose of the smoothed spectrum on the frequency-axis
    low = np.zeros((filter_freqs.shape[0], nseg))
    high = np.zeros((filter_freqs.shape[0], nseg))

    # Index of the lower and higher limit of each frequency bin into the original spectrum
    for i in range(len(filter_freqs)):
        low[i,:] = np.argmin(np.abs(freqs_out - filter_freqs[i, 0]))
        high[i,:] = np.argmin(np.abs(freqs_out - filter_freqs[i, 2]))
    low = low.astype(int)
    high = high.astype(int)

    smooth_spec = np.zeros((nperseg, nseg))
    for i in range(nseg):
        for j in range(filter_freqs.shape[0]):
            smooth_spec[low[j,i] : high[j,i], i] = smoothed_spectrum[j,i]
    
    if smooth_spec.shape[1] == 1:
        smooth_spec = np.squeeze(smooth_spec)

    return smooth_spec



# -*- coding: utf-8 -*-

# Standard library import
from numpy import array, concatenate, zeros, log10, power, argmin, split, arange, interp, iscomplex

def freq_band_synthesis(spectrum, freqs, fmin, fmax):
    """Adapt input spectrum to frequency band levels
    
    Convert the input spectrum to frequency band spectrum
    between "fmin" and "fmax".
    
    Parameters
    ----------
    spectrum : numpy.ndarray
        One-sided spectrum of the signal in [dB], size (nperseg, nseg).
    freqs : list
        List of input frequency , size (nperseg) or (nperseg, nseg).
    fmin : float
        Min frequency band [Hz].
    fmax : float
        Max frequency band [Hz].
    n : int
        Number of bands pr octave.
    G : int
        System for specifying the exact geometric mean frequencies.
        Can be base 2 or base 10.
    fr : int
        Reference frequency. Shall be set to 1 kHz for audible frequency
        range, to 1 Hz for infrasonic range (f < 20 Hz) and to 1 MHz for
        ultrasonic range (f > 31.5 kHz).
    Outputs
    -------
    spec : numpy.ndarray
        Third octave band spectrum of signal sig [dB re.2e-5 Pa], size (nbands, nseg).
    fpref : numpy.ndarray
        Corresponding preferred third octave band center frequencies, size (nbands).
    """
    if iscomplex(spectrum).any():
        raise ValueError('Input spectrum must be in dB, no complex value allowed.')
    
    if (fmin.min() < freqs.min()):
        print("[WARNING] Input spectrum minimum frequency if higher than fmin. Empty values will be filled with 0.")
        df = freqs[1] - freqs[0]
        spectrum = interp(arange(fmin.min(),fmax.max()+df, df), freqs, spectrum)
        freqs = arange(fmin.min(),fmax.max()+df, df)

    if (fmax.max() > freqs.max()):
        print("[WARNING] Input spectrum maximum frequency if lower than fmax. Empty values will be filled with 0.")
        df = freqs[1] - freqs[0]
        spectrum = interp(arange(fmin.min(),fmax.max()+df, df), freqs, spectrum)
        freqs = arange(fmin.min(),fmax.max()+df, df)
    
    # Find the lower and upper index of each band
    idx_low = argmin(abs(freqs[:,None] - fmin), axis=0)
    idx_up = argmin(abs(freqs[:,None] - fmax), axis=0)
    idx = concatenate((idx_low, [idx_up[-1]]))
    
    # Split the given spectrum in several bands
    bands = array(split(spectrum, idx), dtype=object)[1:-1]
    
    # Compute the bands level
    band_spectrum = zeros((len(bands)))
    i = 0
    for s in bands:
        band_spectrum[i] = 10*log10(sum(power(10,s/10)))      
        i += 1

    return band_spectrum, (fmin+fmax)/2


        



